from functools import partial
from typing import Callable, Sequence, Tuple
from jax import Array
import jax
import jax.numpy as jnp
import distrax
import flax.linen as nn
from flax.linen.initializers import constant, orthogonal


class MLPEncoder(nn.Module):
    hidden_size: int = 64

    @nn.compact
    def __call__(self, x):
        return nn.Sequential(
            [
                nn.Dense(self.hidden_size),
                nn.tanh,
                nn.Dense(self.hidden_size),
                nn.tanh,
            ]
        )(x)


class ConvEncoder(nn.Module):
    hidden_size: int = 64

    @nn.compact
    def __call__(self, x):
        return nn.Sequential(
            [
                nn.Conv(16, kernel_size=(2, 2)),
                nn.relu,
                nn.Conv(32, kernel_size=(2, 2)),
                nn.relu,
                nn.Conv(64, kernel_size=(2, 2)),
                nn.relu,
                jnp.ravel,
                nn.Dense(self.hidden_size),
                nn.relu,
            ]
        )(x)


class ActorCritic(nn.Module):
    action_dim: int
    actor_encoder: nn.Module = MLPEncoder()
    critic_encoder: nn.Module = MLPEncoder()

    def setup(self):
        self.actor = nn.Sequential(
            [
                self.actor_encoder,
                nn.Dense(
                    self.action_dim,
                    kernel_init=orthogonal(0.01),
                    bias_init=constant(0.0),
                ),
                # lambda x: distrax.Categorical(logits=x),
            ]
        )

        self.critic = nn.Sequential(
            [
                self.critic_encoder,
                nn.Dense(1, kernel_init=orthogonal(1.0), bias_init=constant(0.0)),
                # lambda x: jnp.squeeze(x, axis=-1),
            ]
        )

    def __call__(self, x: Array) -> Tuple[distrax.Distribution, Array]:
        return distrax.Categorical(self.actor(x)), jnp.squeeze(self.critic(x), -1)

    def policy(self, x: Array) -> distrax.Distribution:
        return distrax.Categorical(logits=self.actor(x))

    def value(self, x: Array) -> Array:
        return jnp.squeeze(self.critic(x), -1)



import jax
import jax.numpy as jnp

import navix as nx
from navix.states import State
from navix.entities import Entities, Player, Goal, Key, Door
from navix.components import EMPTY_POCKET_ID
from navix.rendering.cache import RenderingCache, TILE_SIZE
from navix.rendering.registry import SPRITES_REGISTRY, PALETTE


def test_rgb():
    height = 10
    width = 10
    grid = jnp.zeros((height - 2, width - 2), dtype=jnp.int32)
    grid = jnp.pad(grid, 1, mode="constant", constant_values=-1)

    players = Player(
        position=jnp.asarray((1, 1)), direction=jnp.asarray(0), pocket=EMPTY_POCKET_ID
    )
    goals = Goal.create(position=jnp.asarray((4, 4)), probability=jnp.asarray(1.0))
    keys = Key(position=jnp.asarray((2, 2)), id=jnp.asarray(0), colour=PALETTE.YELLOW)
    doors = Door(
        position=jnp.asarray([(1, 5), (1, 6)]),
        requires=jnp.asarray((0, 0)),
        open=jnp.asarray((False, True)),
        colour=PALETTE.YELLOW[None],
    )

    entities = {
        Entities.PLAYER: players[None],
        Entities.GOAL: goals[None],
        Entities.KEY: keys[None],
        Entities.DOOR: doors,
    }

    state = State(
        key=jax.random.PRNGKey(0),
        grid=grid,
        cache=RenderingCache.init(grid),
        entities=entities,
    )
    sprites_registry = SPRITES_REGISTRY

    doors = state.get_doors()
    doors = doors.replace(open=jnp.asarray((False, True)))
    state.entities[Entities.DOOR] = doors

    obs = nx.observations.rgb(state)
    expected_obs_shape = (
        height * TILE_SIZE,
        width * TILE_SIZE,
        3,
    )
    assert (
        obs.shape == expected_obs_shape
    ), f"Expected observation {expected_obs_shape}, got {obs.shape} instead"

    def get_tile(position):
        x = position[0] * TILE_SIZE
        y = position[1] * TILE_SIZE
        return obs[x : x + TILE_SIZE, y : y + TILE_SIZE, :]

    player = state.get_player()
    player_tile = get_tile(player.position)
    assert jnp.array_equal(
        player_tile, sprites_registry[Entities.PLAYER][player.direction]
    ), player_tile

    goals = state.get_goals()
    goal_tile = get_tile(goals.position[0])
    assert jnp.array_equal(goal_tile, sprites_registry[Entities.GOAL]), goal_tile

    keys = state.get_keys()
    key_tile = get_tile(keys.position[0])
    colour = keys.colour[0]
    assert jnp.array_equal(key_tile, sprites_registry[Entities.KEY][colour]), key_tile

    doors = state.get_doors()
    door = doors[0]
    door_tile = get_tile(door.position)
    colour = door.colour
    idx = jnp.asarray(door.open + 2 * door.locked, dtype=jnp.int32)
    assert jnp.array_equal(
        door_tile, sprites_registry[Entities.DOOR][colour, idx]
    ), door_tile

    door = doors[1]
    door_tile = get_tile(door.position)
    colour = door.colour
    idx = jnp.asarray(door.open + 2 * door.locked, dtype=jnp.int32)
    assert jnp.array_equal(
        door_tile, sprites_registry[Entities.DOOR][colour, idx]
    ), door_tile

    return


def test_categorical_first_person():
    height = 10
    width = 10
    grid = jnp.zeros((height - 2, width - 2), dtype=jnp.int32)
    grid = jnp.pad(grid, 1, mode="constant", constant_values=-1)

    players = Player(
        position=jnp.asarray((1, 1)), direction=jnp.asarray(0), pocket=EMPTY_POCKET_ID
    )
    goals = Goal.create(position=jnp.asarray((4, 4)), probability=jnp.asarray(1.0))
    keys = Key(position=jnp.asarray((2, 2)), id=jnp.asarray(0), colour=PALETTE.YELLOW)
    doors = Door(
        position=jnp.asarray([(1, 5), (1, 6)]),
        requires=jnp.asarray((0, 0)),
        open=jnp.asarray((False, True)),
        colour=PALETTE.YELLOW,
    )
    entities = {
        Entities.PLAYER: players[None],
        Entities.GOAL: goals[None],
        Entities.KEY: keys[None],
        Entities.DOOR: doors,
    }

    state = State(
        key=jax.random.PRNGKey(0),
        grid=grid,
        cache=RenderingCache.init(grid),
        entities=entities,
    )

    obs = nx.observations.categorical_first_person(state)
    print(obs)


if __name__ == "__main__":
    test_rgb()
    # test_categorical_first_person()
    # jax.jit(test_categorical_first_person)()



from dataclasses import dataclass
from typing import Tuple
import wandb

import numpy as np
import jax.numpy as jnp
import flax.linen as nn
from flax.linen.initializers import constant, orthogonal
import tyro
import navix as nx
from navix.environments.environment import Environment
from navix.agents import PPO, PPOHparams, ActorCritic


def FlattenObsWrapper(env: Environment):
    flatten_obs_fn = lambda x: jnp.ravel(env.observation_fn(x))
    flatten_obs_shape = (int(np.prod(env.observation_space.shape)),)
    return env.replace(
        observation_fn=flatten_obs_fn,
        observation_space=env.observation_space.replace(shape=flatten_obs_shape),
    )


@dataclass
class Args:
    project_name = "navix-baselines"
    seeds_range: Tuple[int, int, int] = (0, 10, 1)
    ppo: PPOHparams = PPOHparams()


if __name__ == "__main__":
    args = tyro.cli(Args)

    # create environments
    for env_id in nx.registry():
        # init logging
        config = {**vars(args), **{"observations": "symbolic"}, **{"algo": "ppo"}}
        wandb.init(project=args.project_name, config=config)

        # init environment
        env = nx.make(env_id)
        env = FlattenObsWrapper(env)

        # create agent
        agent = PPO(
            hparams=args.ppo,
            network=ActorCritic(action_dim=len(env.action_set)),
            env=env,
        )

        # run experiment
        experiment = nx.Experiment(
            name=args.project_name,
            agent=agent,
            env=env,
            env_id=env_id,
            seeds=tuple(range(*args.seeds_range)),
        )
        experiment.run()



from dataclasses import dataclass
import time
from typing import Dict, Tuple

import numpy as np
import wandb
import jax
import jax.numpy as jnp
from flax import struct
from flax.training.train_state import TrainState


class HParams(struct.PyTreeNode):
    debug: bool = struct.field(pytree_node=False, default=False)
    """Whether to run in debug mode."""
    log_frequency: int = struct.field(pytree_node=False, default=1)
    """How often to log results."""
    log_render: bool = struct.field(pytree_node=False, default=False)


class Agent(struct.PyTreeNode):
    hparams: HParams

    def train(self, rng: jax.Array) -> Tuple[TrainState, Dict[str, jax.Array]]:
        raise NotImplementedError

    def log(self, logs, inspectable=None):
        if len(logs) == 0 or logs["iter/updates"] % self.hparams.log_frequency != 0:
            return

        start_time = time.time()
        msg = f"Update Step: {logs['iter/updates']}, Frames: {logs['iter/frames']}"
        step = jnp.asarray(logs["iter/updates"], dtype=jnp.int32)

        # log renders
        if self.hparams.log_render:
            render_human = logs.pop("render/human")  # (T, 3, H, W)
            logs[f"render/human"] = wandb.Video(np.array(render_human), fps=4)

        if "done_mask" in logs:
            mask = jnp.asarray(logs.pop("done_mask"), dtype=jnp.bool)  # (T, N)
            # log episode length
            if "lengths" in logs:
                lengths: jax.Array = logs.pop("lengths")  # (T, N)
                episode_lengths = lengths[mask]  # (K,)
                logs["perf/episode_length"] = jnp.mean(episode_lengths)
                msg += f", Length: {logs['perf/episode_length']}"

            # log returns
            if "returns" in logs:
                returns = logs.pop("returns")  # (T, N)
                final_returns = returns[mask]  # (K,)
                logs["perf/returns"] = jnp.mean(final_returns)
                logs["perf/success_rate"] = jnp.mean(final_returns == 1.0)
                msg += f", Returns: {logs['perf/returns']}, Success Rate: {logs['perf/success_rate']}"

        msg += f", Logging time cost: {time.time() - start_time}"
        wandb.log(logs, step=step)

    def log_on_train_end(self, logs):
        print(jax.tree.map(lambda x: x.shape, logs))
        len_logs = len(logs["iter/updates"])
        for step in range(len_logs):
            step_logs = {k: jax.tree.map(lambda x: x[step], v) for k, v in logs.items()}
            self.log(step_logs)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

from typing import Dict, Tuple

import jax
from jax import Array
import jax.numpy as jnp
from flax import struct

from .registry import TILE_SIZE, SPRITES_REGISTRY


class RenderingCache(struct.PyTreeNode):
    patches: Array
    """A flat set of patches representing the RGB values of each tile in the base map"""

    @classmethod
    def init(cls, grid: Array) -> RenderingCache:
        background = render_background(grid)
        patches = flatten_patches(background)

        # add discard pile
        patches = jnp.concatenate(
            [
                patches,
                jnp.zeros((1, TILE_SIZE, TILE_SIZE, 3), dtype=jnp.uint8),
            ],
            axis=0,
        )
        return cls(patches=patches)


def render_background(
    grid: Array, sprites_registry: Dict[str, Array] = SPRITES_REGISTRY
) -> Array:
    image_width = grid.shape[0] * TILE_SIZE
    image_height = grid.shape[1] * TILE_SIZE
    n_channels = 3

    background = jnp.zeros((image_height, image_width, n_channels), dtype=jnp.uint8)
    grid_resized = jax.image.resize(
        grid, (grid.shape[0] * TILE_SIZE, grid.shape[1] * TILE_SIZE), method="nearest"
    )

    mask = jnp.asarray(grid_resized, dtype=bool)  # 0 = floor, 1 = wall
    # index by [entity_type, direction, open/closed, y, x, channel]
    wall_tile = tile_grid(grid, sprites_registry["wall"])
    floor_tile = tile_grid(grid, sprites_registry["floor"])
    background = jnp.where(mask[..., None], wall_tile, floor_tile)
    return background


def tile_grid(grid: Array, tile: Array) -> Array:
    """Tiles a grid (H, W) with equal tiles `tiles` (w, h, 3) to get a final array
    of shape (H * h, W * w, 3) and dtype `jnp.uint8`"""
    tiled = jnp.tile(tile, (*grid.shape, 1))
    return jnp.asarray(tiled, dtype=jnp.uint8)


def flatten_patches(
    image: Array, patch_size: Tuple[int, int] = (TILE_SIZE, TILE_SIZE)
) -> Array:
    height = image.shape[0] // patch_size[0]
    width = image.shape[1] // patch_size[1]
    n_channels = image.shape[2]

    grid = image.reshape(height, patch_size[0], width, patch_size[1], n_channels)

    # Swap the first and second axes of the grid to revert the stacking order
    grid = jnp.swapaxes(grid, 1, 2)

    # Reshape the grid of tiles into the original list of tiles
    patches = grid.reshape(height * width, patch_size[0], patch_size[1], n_channels)

    return patches


def unflatten_patches(patches: Array, image_size: Tuple[int, int]) -> Array:
    image_height = image_size[0]
    image_width = image_size[1]
    patch_height = patches.shape[1]
    patch_width = patches.shape[2]
    n_channels = patches.shape[3]

    # Reshape the list of tiles into a 2D grid
    grid = patches.reshape(
        image_height // patch_height,
        image_width // patch_width,
        patch_height,
        patch_width,
        n_channels,
    )

    # Swap the first and second axes of the grid to change the order of stacking
    grid = jnp.swapaxes(grid, 1, 2)

    # Reshape and stack the grid tiles horizontally and vertically to form the final image
    image = grid.reshape(image_height, image_width, n_channels)

    return image



import jax
import jax.numpy as jnp
import navix as nx


def test_room():
    def f():
        env = nx.environments.Room.create(height=3, width=3, max_steps=8)
        key = jax.random.PRNGKey(4)
        reset = jax.jit(env._reset)
        step = jax.jit(env.step)
        timestep = reset(key)
        # these are optimal actios for navigation + action_cost
        actions = (
            0,  # noop sanity check
            2,  # rotate_ccw
            3,  # forward
            3,  # forward
            2,  # rotate_ccw
            3,  # forward
        )
        print(timestep)
        print()
        for action in actions:
            timestep = step(timestep, jnp.asarray(action))
            print()
            print(nx.actions.DEFAULT_ACTION_SET[action])
            print(timestep)
        return timestep

    f()
    timestep = jax.jit(f)()
    print(timestep)


def test_keydoor():
    def f():
        env = nx.environments.DoorKey.create(height=5, width=10, max_steps=8)
        key = jax.random.PRNGKey(1)
        reset = jax.jit(env._reset)
        step = jax.jit(env.step)
        timestep = reset(key)
        #  these are optimal actions for navigation + action_cost
        actions = (
            0,  # rotate_ccw
            2,  # forward
            2,  # forward
            2,  # forward
            0,  # rotate_ccw
            3,  # pick-up
            0,  # rotate_ccw
            0,  # rotate_ccw
            2,  # forward
            2,  # forward
            1,  # rotate_cw
            2,  # forward
            0,  # rotate_ccw
            5,  # open
            2,  # forward
            2,  # forward
        )
        print(timestep)
        for action in actions:
            timestep = step(timestep, jnp.asarray(action))
            print()
            print(nx.actions.DEFAULT_ACTION_SET[action])
            print(timestep)
        return timestep

    f()
    jax.jit(f)()


def test_keydoor2():
    env = nx.environments.DoorKey.create(5, 7, 100, observation_fn=nx.observations.rgb)

    key = jax.random.PRNGKey(1)
    timestep = env._reset(key)
    return


if __name__ == "__main__":
    # test_room()
    # jax.jit(test_room)()
    test_keydoor()
    # test_keydoor2()



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

from typing import Union
import jax
import jax.numpy as jnp
from jax import Array
import jax.tree_util as jtu

from navix import observations, rewards, terminations

from ..components import EMPTY_POCKET_ID
from ..rendering.cache import RenderingCache
from ..environments import Environment
from ..entities import Ball, Player, Key, Door
from ..states import State
from ..environments import Timestep
from ..grid import random_directions, random_colour, RoomsGrid
from .registry import register_env


class KeyCorridor(Environment):
    def _reset(self, key: Array, cache: Union[RenderingCache, None] = None) -> Timestep:
        n_rows_config = {3: 1, 5: 2}
        n_rows = n_rows_config.get(self.height, 3)
        room_size = (self.width - 3) // 3
        k1, k2, k3, k4, k5, k6 = jax.random.split(key, num=6)

        # grid of rooms
        grid = RoomsGrid.create(n_rows, 3, (room_size, room_size))

        # key
        key_room_row = jax.random.randint(k1, (), minval=0, maxval=n_rows)
        key_pos = grid.position_in_room(
            key_room_row, jnp.asarray(0, dtype=jnp.int32), key=k1
        )
        key_colour = random_colour(k4)
        key_id = jnp.asarray(1)
        key_obj = Key.create(key_pos, key_colour, key_id)

        # agent
        pk_1, pk_2, pk_3 = jax.random.split(k2, num=3)
        agent_room_row = jax.random.randint(pk_1, (), minval=0, maxval=n_rows)
        agent_pos = grid.position_in_room(agent_room_row, jnp.asarray(1), key=pk_2)
        player = Player.create(
            agent_pos, random_directions(pk_3), pocket=EMPTY_POCKET_ID
        )

        # ball
        ball_room_row = jax.random.randint(k3, (), minval=0, maxval=n_rows)
        ball_pos = grid.position_in_room(ball_room_row, jnp.asarray(2), key=k4)
        ball = Ball.create(ball_pos, random_colour(k6), probability=jnp.asarray(0.0))

        # Doors
        doors = []
        for row in range(n_rows):
            k5, k6, k7, k8, k9 = jax.random.split(k5, num=5)
            # left corridor, right wall
            door_pos = grid.position_on_border(row, 2, 0, key=k5)
            requires, colour, open = jax.lax.cond(
                jnp.array_equal(row, ball_room_row),
                lambda: (key_id, key_colour, jnp.asarray(2)),
                lambda: (jnp.asarray(-1), random_colour(k5), jnp.asarray(0)),
            )
            doors.append(
                Door.create(
                    position=door_pos, requires=requires, colour=colour, open=open
                )
            )
            # right corridor, left wall
            door_pos = grid.position_on_border(row, 0, 1, key=k7)
            doors.append(
                Door.create(
                    position=door_pos,
                    requires=EMPTY_POCKET_ID,
                    colour=random_colour(k7),
                    open=jnp.asarray(0),
                )
            )
        for row in range(n_rows - 1):
            k9, k10, k11, k12 = jax.random.split(k9, num=4)
            # first col
            door_pos = grid.position_on_border(row, 0, 3, key=k9)
            doors.append(
                Door.create(
                    position=door_pos,
                    requires=EMPTY_POCKET_ID,
                    colour=random_colour(k10),
                    open=jnp.asarray(0),
                )
            )
            doors.append(
                Door.create(
                    position=door_pos,
                    requires=EMPTY_POCKET_ID,
                    colour=random_colour(k12),
                    open=jnp.asarray(0),
                )
            )
        doors = jtu.tree_map(lambda *x: jnp.stack(x), *doors)

        entities = {
            "player": player[None],
            "key": key_obj[None],
            "door": doors,
            "goal": ball[None],
        }

        grid = grid.get_grid()
        grid = grid.at[
            1 + room_size : self.height - 1 : room_size + 1,
            1 + room_size + 1 : 1 + room_size + 1 + room_size,
        ].set(0)
        state = State(
            key=key,
            grid=grid,
            cache=cache or RenderingCache.init(grid),
            entities=entities,
        )
        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation_fn(state),
            action=jnp.asarray(-1, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )


register_env(
    "Navix-KeyCorridorS3R1-v0",
    lambda *args, **kwargs: KeyCorridor.create(
        *args,
        **kwargs,
        height=3,
        width=7,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
    ),
)
register_env(
    "Navix-KeyCorridorS3R2-v0",
    lambda *args, **kwargs: KeyCorridor.create(
        *args,
        **kwargs,
        height=5,
        width=7,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
    ),
)
register_env(
    "Navix-KeyCorridorS3R3-v0",
    lambda *args, **kwargs: KeyCorridor.create(
        *args,
        **kwargs,
        height=7,
        width=7,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
    ),
)
register_env(
    "Navix-KeyCorridorS4R3-v0",
    lambda *args, **kwargs: KeyCorridor.create(
        *args,
        **kwargs,
        height=10,
        width=10,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
    ),
)
register_env(
    "Navix-KeyCorridorS5R3-v0",
    lambda *args, **kwargs: KeyCorridor.create(
        *args,
        **kwargs,
        height=13,
        width=13,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
    ),
)
register_env(
    "Navix-KeyCorridorS6R3-v0",
    lambda *args, **kwargs: KeyCorridor.create(
        *args,
        **kwargs,
        height=16,
        width=16,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
    ),
)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations
from typing import Union

import jax.numpy as jnp
from jax import Array
from flax import struct

from navix import observations, rewards, terminations

from ..components import EMPTY_POCKET_ID
from ..entities import Entities, Goal, Lava, Player
from ..states import State
from ..grid import room
from ..rendering.cache import RenderingCache
from .environment import Environment, Timestep
from .registry import register_env


class DistShift(Environment):
    split_lava: bool = struct.field(pytree_node=False, default=False)

    def _reset(self, key: Array, cache: Union[RenderingCache, None] = None) -> Timestep:
        # map
        grid = room(height=self.height, width=self.width)

        # goal and player
        player_pos = jnp.asarray([1, 1])
        direction = jnp.asarray(0)
        player = Player.create(
            position=player_pos,
            direction=direction,
            pocket=EMPTY_POCKET_ID,
        )
        # goal
        goal_pos = jnp.asarray([1, self.width - 2])
        goal = Goal.create(position=goal_pos, probability=jnp.asarray(1.0))

        # lava
        last_row = 5 if self.split_lava else 2
        lava_pos = jnp.asarray(
            [[1, 3], [1, 4], [1, 5], [last_row, 3], [last_row, 4], [last_row, 5]]
        )
        lava = Lava.create(lava_pos)

        entities = {
            Entities.PLAYER: player[None],
            Entities.GOAL: goal[None],
            Entities.LAVA: lava,
        }

        # systems
        state = State(
            key=key,
            grid=grid,
            cache=cache or RenderingCache.init(grid),
            entities=entities,
        )

        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation_fn(state),
            action=jnp.asarray(0, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )


register_env(
    "Navix-DistShift1-v0",
    lambda *args, **kwargs: DistShift.create(
        height=7,
        width=9,
        split_lava=False,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-DistShift2-v0",
    lambda *args, **kwargs: DistShift.create(
        height=7,
        width=9,
        split_lava=True,
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),        
        *args,
        **kwargs,
    ),
)



from typing import Any


class Config:
    """Config class to store global variables."""
    
    def __init__(self):
        self.ARRAY_CHECKS_ENABLED = False

    def update(self, key: str, value: Any) -> None:
        setattr(self, key, value)

    def reset(self) -> None:
        self.__init__()


config = Config()



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

from typing import Callable, Tuple
from jax import Array
import jax
import jax.numpy as jnp
from .entities import Entities, Ball
from .states import EventsManager, State
from .grid import positions_equal, translate


def deterministic_transition(
    state: State, action: Array, actions_set: Tuple[Callable[[State], State], ...]
) -> State:
    """Deterministic transition function. It selects the action from the set of actions
    and applies it to the state.
    
    Args:
        state (State): The current state of the game.
        action (Array): The action to be taken.
        actions_set (Tuple[Callable[[State], State]): A set of actions that can be taken.
    
    Returns:
        State: The new state of the game."""
    return jax.lax.switch(action, actions_set, state)


def stochastic_transition(
    state: State, action: Array, actions_set: Tuple[Callable[[State], State], ...]
) -> State:
    """Stochastic transition function. It selects the action from the set of actions
    and applies it to the state, and updates entities that have stochastic transitions,
    such as balls.
    
    Args:
        state (State): The current state of the game.
        action (Array): The action to be taken.
        actions_set (Tuple[Callable[[State], State]): A set of actions that can be taken.
    
    Returns:
        State: The new state of the game."""
    # actions
    state = jax.lax.switch(action, actions_set, state)

    state = update_balls(state)
    return state


def update_balls(state: State) -> State:
    """Update the position of the balls in the game.
    Balls move in a random direction if they can, otherwise they stay in place.
    
    Args:
        state (State): The current state of the game.
    
    Returns:
        State: The new state of the game."""
    def update_one(ball: Ball, key: Array) -> Tuple[Array, EventsManager]:
        direction = jax.random.randint(key, (), minval=0, maxval=4)
        new_position = translate(ball.position, direction)
        new_ball = ball.replace(position=new_position)
        can_move, events = _can_spawn_there(state, new_ball)
        return jnp.where(can_move, new_ball.position, ball.position), events

    if Entities.BALL in state.entities:
        balls: Ball = state.entities[Entities.BALL]  # type: ignore
        keys = jax.random.split(state.key, len(balls.position) + 1)
        new_position, new_events = jax.jit(jax.vmap(update_one))(balls, keys[1:])
        # update balls
        balls = balls.replace(position=new_position)
        state = state.set_balls(balls)
        # update events
        # take only the first happened event (even if happened already)
        idx = jnp.where(new_events.ball_hit.happened, size=1)[0][0]  # scalar
        ball_hits = jax.tree.map(lambda x: x[idx], new_events.ball_hit)
        events = state.events.replace(ball_hit=ball_hits)
        state = state.replace(key=keys[0], events=events)
    return state


def _can_spawn_there(state: State, ball: Ball) -> Tuple[Array, EventsManager]:
    # according to the grid
    walkable = jnp.equal(state.grid[tuple(ball.position)], 0)

    # according to entities
    events = state.events
    entities = state.entities
    for k in state.entities:
        obstructs = positions_equal(entities[k].position, ball.position)[0]
        if k == Entities.PLAYER:
            events = jax.lax.cond(
                obstructs,
                lambda x: x.record_ball_hit(ball),
                lambda x: x,
                events,
            )
        walkable = jnp.logical_and(walkable, jnp.any(jnp.logical_not(obstructs)))
    return jnp.asarray(walkable, dtype=jnp.bool_), events


DEFAULT_TRANSITION = stochastic_transition



import jax
import jax.numpy as jnp

import navix as nx
from navix.states import State
from navix.entities import Entities, Player, Goal, Key, Door
from navix.components import EMPTY_POCKET_ID
from navix.rendering.registry import PALETTE


def test_navigation():
    """Unittest for https://github.com/epignatelli/navix/pull/47"""
    height = 10
    width = 10
    grid = jnp.zeros((height - 2, width - 2), dtype=jnp.int32)
    grid = jnp.pad(grid, 1, mode="constant", constant_values=-1)

    players = Player(
        position=jnp.asarray((1, 1)), direction=jnp.asarray(0), pocket=EMPTY_POCKET_ID
    )
    goals = Goal.create(
        position=jnp.asarray([(1, 1), (1, 1)]), probability=jnp.asarray([0.0, 0.0])
    )
    keys = Key.create(position=jnp.asarray((2, 2)), id=jnp.asarray(0), colour=PALETTE.YELLOW)
    doors = Door.create(
        position=jnp.asarray([(1, 5), (1, 6)]),
        requires=jnp.asarray((0, 0)),
        open=jnp.asarray((False, True)),
        colour=PALETTE.YELLOW,
    )

    entities = {
        Entities.PLAYER: players[None],
        Entities.GOAL: goals,
        Entities.KEY: keys[None],
        Entities.DOOR: doors,
    }

    state = State(
        key=jax.random.PRNGKey(0),
        grid=grid,
        cache=nx.rendering.cache.RenderingCache.init(grid),
        entities=entities,
    )
    action = jnp.asarray(0)
    reward = nx.rewards.on_goal_reached(state, action, state)
    assert jnp.array_equal(reward, jnp.asarray(0.0))


def test_tasks_composition():
    reward_fn = nx.rewards.compose(
        nx.rewards.on_goal_reached,
        nx.rewards.action_cost,
        nx.rewards.time_cost,
        nx.rewards.wall_hit_cost,
    )

    env = nx.environments.Room.create(height=3, width=3, max_steps=8, reward_fn=reward_fn)
    key = jax.random.PRNGKey(0)

    def _test():
        timestep = env._reset(key)
        for _ in range(10):
            timestep = env.step(timestep, jax.random.randint(key, (), 0, 7))
        return timestep

    print(jax.jit(_test)())


if __name__ == "__main__":
    # test_tasks_composition()
    test_navigation()



# docs and experiment results can be found at https://docs.cleanrl.dev/rl-algorithms/ppo/#ppo_procgenpy
from __future__ import annotations
from functools import reduce
import operator
import random
import time
from dataclasses import dataclass

from minigrid.wrappers import ImgObsWrapper, FullyObsWrapper
import gymnasium as gym
from gymnasium.wrappers.record_episode_statistics import RecordEpisodeStatistics
from gymnasium.wrappers.record_video import RecordVideo
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import tyro
from torch.distributions.categorical import Categorical
import wandb


@dataclass
class Args:
    exp_name: str = "navix-benchmarks"
    """the name of this experiment"""
    seed: int = 0
    """seed of the experiment"""
    torch_deterministic: bool = True
    """if toggled, `torch.backends.cudnn.deterministic=False`"""
    cuda: bool = True
    """if toggled, cuda will be enabled by default"""
    capture_video: bool = False
    """whether to capture videos of the agent performances (check out `videos` folder)"""

    # Algorithm specific arguments
    env_id: str = "MiniGrid-DoorKey-8x8-v0"
    """the id of the environment"""
    total_timesteps: int = 10_000_000
    """total timesteps of the experiments"""
    learning_rate: float = 2.5e-4
    """the learning rate of the optimizer"""
    num_envs: int = 16
    """the number of parallel game environments"""
    num_steps: int = 128
    """the number of steps to run in each environment per policy rollout"""
    anneal_lr: bool = True
    """Toggle learning rate annealing for policy and value networks"""
    gamma: float = 0.99
    """the discount factor gamma"""
    gae_lambda: float = 0.95
    """the lambda for the general advantage estimation"""
    num_minibatches: int = 8
    """the number of mini-batches"""
    update_epochs: int = 1
    """the K epochs to update the policy"""
    norm_adv: bool = True
    """Toggles advantages normalization"""
    clip_coef: float = 0.2
    """the surrogate clipping coefficient"""
    clip_vloss: bool = True
    """Toggles whether or not to use a clipped loss for the value function, as per the paper."""
    ent_coef: float = 0.01
    """coefficient of the entropy"""
    vf_coef: float = 0.5
    """coefficient of the value function"""
    max_grad_norm: float = 0.5
    """the maximum norm for the gradient clipping"""
    target_kl: float | None = None
    """the target KL divergence threshold"""

    # to be filled in runtime
    batch_size: int = 0
    """the batch size (computed in runtime)"""
    minibatch_size: int = 0
    """the mini-batch size (computed in runtime)"""
    num_iterations: int = 0
    """the number of iterations (computed in runtime)"""
    annotations_path: str = ""
    """the path to the LLM annotations file"""


def layer_init(layer, std=np.sqrt(2), bias_const=0.0):
    torch.nn.init.orthogonal_(layer.weight, std)
    torch.nn.init.constant_(layer.bias, bias_const)
    return layer


class MLPEncoder(nn.Module):
    def __init__(self, in_dim, hidden_size=64):
        super().__init__()
        self.network = nn.Sequential(
            nn.Linear(in_dim, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
        )

    def forward(self, x):
        return self.network(x)


class Agent(nn.Module):
    def __init__(self, encoder: nn.Module, envs, hidden_size=64):
        super().__init__()
        self.network = encoder
        self.actor = layer_init(
            nn.Linear(hidden_size, envs.single_action_space.n), std=0.01
        )
        self.critic = layer_init(nn.Linear(hidden_size, 1), std=1)

    def get_value(self, x):
        return self.critic(self.network(x))

    def get_action_and_value(self, x, action=None):
        hidden = self.network(x)  # "bhwc" -> "bchw"
        logits = self.actor(hidden)
        probs = Categorical(logits=logits)
        if action is None:
            action = probs.sample()
        return (
            action,
            probs.log_prob(action),
            probs.entropy(),
            self.critic(hidden),
        )


class FlattenObsWrapper(gym.ObservationWrapper):
    def __init__(self, env):
        super().__init__(env)
        self.observation_space = gym.spaces.Box(
            low=0.0,
            high=1.0,
            shape=(reduce(operator.mul, env.observation_space.shape, 1),),  # type: ignore
            dtype=np.float32,
        )

    def observation(self, obs):
        return obs.flatten()


if __name__ == "__main__":
    args = tyro.cli(Args)
    args.batch_size = int(args.num_envs * args.num_steps)
    args.minibatch_size = int(args.batch_size // args.num_minibatches)
    args.num_iterations = args.total_timesteps // args.batch_size
    run_name = f"{args.env_id}__{args.exp_name}__{args.seed}__{int(time.time())}"
    wandb.init(
        project=args.exp_name,
        group="minigrid",
        config=vars(args),
        name=run_name,
        monitor_gym=True,
        save_code=True,
    )

    # TRY NOT TO MODIFY: seeding
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    torch.backends.cudnn.deterministic = args.torch_deterministic

    device = torch.device("cuda" if torch.cuda.is_available() and args.cuda else "cpu")

    # Load environments
    envs = gym.vector.make(
        args.env_id,
        max_episode_steps=100,
        num_envs=args.num_envs,
        asynchronous=True,
        wrappers=[FullyObsWrapper, ImgObsWrapper, FlattenObsWrapper],
    )
    envs.is_vector_env = True
    envs = RecordEpisodeStatistics(envs)
    if args.capture_video:
        envs = RecordVideo(envs, f"videos/{run_name}")
    assert isinstance(
        envs.single_action_space, gym.spaces.Discrete
    ), "only discrete action space is supported"

    encoder = MLPEncoder(envs.single_observation_space.shape[0])
    agent = Agent(encoder, envs).to(device)
    optimizer = optim.Adam(agent.parameters(), lr=args.learning_rate, eps=1e-5)

    # TRAINING LOOP

    # ALGO Logic: Storage setup
    obs = torch.zeros(
        (args.num_steps, args.num_envs) + envs.single_observation_space.shape
    ).to(device)
    action_size: int = envs.single_action_space.n  # type: ignore
    actions = torch.zeros(
        (
            args.num_steps,
            args.num_envs,
        ),
        dtype=torch.int32,
    ).to(device)
    logprobs = torch.zeros((args.num_steps, args.num_envs), dtype=torch.float32).to(
        device
    )
    rewards = torch.zeros((args.num_steps, args.num_envs), dtype=torch.float32).to(
        device
    )
    dones = torch.zeros((args.num_steps, args.num_envs), dtype=torch.bool).to(device)
    values = torch.zeros((args.num_steps, args.num_envs), dtype=torch.float32).to(
        device
    )

    # TRY NOT TO MODIFY: start the game
    frames = 0
    reset_obs, _ = envs.reset()
    next_obs = torch.Tensor(reset_obs).to(device)
    next_done = torch.zeros(args.num_envs).to(device)

    training_time = time.time()

    for iteration in range(1, args.num_iterations + 1):
        start_time = time.time()
        logs = {}
        # Annealing the rate if instructed to do so.
        if args.anneal_lr:
            frac = 1.0 - (iteration - 1.0) / args.num_iterations
            lrnow = frac * args.learning_rate
            optimizer.param_groups[0]["lr"] = lrnow

        for step in range(0, args.num_steps):
            frames += args.num_envs
            obs[step] = next_obs
            dones[step] = next_done

            # ALGO LOGIC: action logic
            with torch.no_grad():
                action, logprob, _, value = agent.get_action_and_value(next_obs)
                values[step] = value.flatten()
            actions[step] = action
            logprobs[step] = logprob

            # TRY NOT TO MODIFY: execute the game and log data.
            next_obs, reward, next_terminated, next_truncated, info = envs.step(
                action.cpu().numpy()
            )
            reward = torch.tensor(reward)
            rewards[step] = reward.to(device).view(-1)
            next_obs = torch.Tensor(next_obs).to(device)
            next_done = torch.Tensor(np.logical_or(next_terminated, next_truncated)).to(
                device
            )

            # for item in info:
            if "episode" in info:
                env_mask = info["_episode"]  # mask for terminal episodes
                returns = info["episode"]["r"][env_mask].mean()
                length = info["episode"]["l"][env_mask].mean()
                is_success = reward[env_mask] >= 1.0
                logs["perf/returns"] = returns
                logs["perf/episode_length"] = length
                logs["perf/success_hits"] = is_success.int().sum()
                logs["perf/success_rate"] = is_success.float().mean()
                print(
                    f"global_step={frames}, episodic_return={returns}", 
                    f"episode_length={length}"
                )

        # bootstrap value if not done
        with torch.no_grad():
            next_value = agent.get_value(next_obs).reshape(1, -1)
            advantages = torch.zeros_like(rewards).to(device)
            lastgaelam = 0
            for t in reversed(range(args.num_steps)):
                if t == args.num_steps - 1:
                    nextnonterminal = 1.0 - next_done
                    nextvalues = next_value
                else:
                    nextnonterminal = 1.0 - dones[t + 1].float()
                    nextvalues = values[t + 1]
                delta = (
                    rewards[t] + args.gamma * nextvalues * nextnonterminal - values[t]
                )
                advantages[t] = lastgaelam = (
                    delta + args.gamma * args.gae_lambda * nextnonterminal * lastgaelam
                )
            returns = advantages + values

        # flatten the batch
        b_obs = obs.reshape((-1,) + envs.single_observation_space.shape)
        b_logprobs = logprobs.reshape(-1)
        b_actions = actions.reshape((-1,))
        b_advantages = advantages.reshape(-1)
        b_returns = returns.reshape(-1)
        b_values = values.reshape(-1)

        # Optimizing the policy and value network
        b_inds = np.arange(args.batch_size)

        clipfracs = []
        for epoch in range(args.update_epochs):
            np.random.shuffle(b_inds)
            for start in range(0, args.batch_size, args.minibatch_size):
                end = start + args.minibatch_size
                mb_inds = b_inds[start:end]

                _, newlogprob, entropy, newvalue = agent.get_action_and_value(
                    b_obs[mb_inds], b_actions.long()[mb_inds]
                )
                logratio = newlogprob - b_logprobs[mb_inds]
                ratio = logratio.exp()

                with torch.no_grad():
                    # calculate approx_kl http://joschu.net/blog/kl-approx.html
                    old_approx_kl = (-logratio).mean()
                    approx_kl = ((ratio - 1) - logratio).mean()
                    clipfracs += [
                        ((ratio - 1.0).abs() > args.clip_coef).float().mean().item()
                    ]

                mb_advantages = b_advantages[mb_inds]
                if args.norm_adv:
                    mb_advantages = (mb_advantages - mb_advantages.mean()) / (
                        mb_advantages.std() + 1e-8
                    )

                # Policy loss
                pg_loss1 = -mb_advantages * ratio
                pg_loss2 = -mb_advantages * torch.clamp(
                    ratio, 1 - args.clip_coef, 1 + args.clip_coef
                )
                pg_loss = torch.max(pg_loss1, pg_loss2).mean()

                # Value loss
                newvalue = newvalue.view(-1)
                if args.clip_vloss:
                    v_loss_unclipped = (newvalue - b_returns[mb_inds]) ** 2
                    v_clipped = b_values[mb_inds] + torch.clamp(
                        newvalue - b_values[mb_inds],
                        -args.clip_coef,
                        args.clip_coef,
                    )
                    v_loss_clipped = (v_clipped - b_returns[mb_inds]) ** 2
                    v_loss_max = torch.max(v_loss_unclipped, v_loss_clipped)
                    v_loss = 0.5 * v_loss_max.mean()
                else:
                    v_loss = 0.5 * ((newvalue - b_returns[mb_inds]) ** 2).mean()

                entropy_loss = entropy.mean()
                loss = pg_loss - args.ent_coef * entropy_loss + v_loss * args.vf_coef

                optimizer.zero_grad()
                loss.backward()
                nn.utils.clip_grad_norm_(agent.parameters(), args.max_grad_norm)
                optimizer.step()

            if args.target_kl is not None and approx_kl > args.target_kl:
                break

        y_pred, y_true = b_values.cpu().numpy(), b_returns.cpu().numpy()
        var_y = np.var(y_true)
        explained_var = np.nan if var_y == 0 else 1 - np.var(y_true - y_pred) / var_y

        # TRY NOT TO MODIFY: record rewards for plotting purposes
        logs["ppo/total_loss"] = loss.item()
        logs["ppo/value_loss"] = v_loss.item()
        logs["ppo/actor_loss"] = pg_loss.item()
        logs["ppo/entropy"] = entropy_loss.item()
        logs["ppo/approx_kl"] = approx_kl.item()
        logs["ppo/clipfrac"] = np.mean(clipfracs)

        logs["iter/frames"] = frames
        logs["iter/epochs"] = iteration * args.update_epochs
        logs["iter/updates"] = iteration * args.update_epochs * args.num_minibatches
        logs["iter/learing_rate"] = optimizer.param_groups[0]["lr"]

        logs["iter/fps"] = int(b_inds.size / (time.time() - start_time))
        wandb.log(logs)

    envs.close()
    print(f"Training completed in {time.time() - training_time} seconds")
    wandb.finish()



import sys

import jax
import jax.numpy as jnp
from navix.spaces import Continuous, Discrete


MAX_INT = 100_000_000
MIN_INT = -100_000_000


def test_discrete():
    key = jax.random.PRNGKey(42)
    elements = (5, 0, MAX_INT, MIN_INT)
    shapes = ((), (0,), (0, 0), (1, 2), (5, 5))
    dtypes = (jnp.int8, jnp.int16, jnp.int32)
    for element in elements:
        for shape in shapes:
            for dtype in dtypes:
                space = Discrete.create(element, shape, dtype)
                sample = space.sample(key)
                print(sample)
                assert jnp.all(jnp.logical_not(jnp.isnan(sample)))


def test_continuous():
    key = jax.random.PRNGKey(42)
    shapes = ((), (0,), (0, 0), (1, 2), (5, 5))
    min_max = [
        (0.0, 1.0),
        (0.0, 1),
        (0, 1),
        (1.0, -1.0),
        (MIN_INT, MAX_INT),
    ]
    for shape in shapes:
        for minimum, maximum in min_max:
            space = Continuous.create(
                shape=shape, minimum=jnp.asarray(minimum), maximum=jnp.asarray(maximum)
            )
            sample = space.sample(key)
            print(sample)
            assert jnp.all(jnp.logical_not(jnp.isnan(sample)))


if __name__ == "__main__":
    test_discrete()
    test_continuous()



import jax
import jax.numpy as jnp

from navix.entities import Goal, Player
from navix.rendering.registry import TILE_SIZE


def test_indexing():
    # batched entity with batch size 1
    entity = Player(
        position=jnp.ones((1, 2), dtype=jnp.int32),
        direction=jnp.ones((1,), jnp.int32),
        pocket=jnp.ones((1,), jnp.int32),
    )
    assert jnp.array_equal(entity[0].position, jnp.asarray((1, 1)))
    assert jnp.array_equal(entity[0].direction, jnp.asarray(1))


def test_get_sprites():
    # batched entity with batch size 1
    entity = Goal.create(position=jnp.ones((1, 2)), probability=jnp.ones((1,)))
    assert entity.sprite.shape == (1, TILE_SIZE, TILE_SIZE, 3)

    # batched entity with batch size > 1
    entity = Goal.create(position=jnp.ones((5, 2)), probability=jnp.ones((5,)))
    assert entity.sprite.shape == (5, TILE_SIZE, TILE_SIZE, 3)


if __name__ == "__main__":
    test_indexing()
    # test_get_sprites()
    jax.jit(test_get_sprites)()



from dataclasses import asdict, replace, fields
import time
from typing import Dict, Tuple

import distrax
import jax
import jax.numpy as jnp
import wandb
import wandb.util
from navix.agents.agent import Agent
from navix.environments.environment import Environment


class Experiment:
    """A class to run an experiment with a given agent and environment.

    Args:
        name (str): The name of the experiment.
        agent (Agent): The agent to use in the experiment.
        env (Environment): The environment to use in the experiment.
        env_id (str): The ID of the environment.
        seeds (Tuple[int, ...]): The seeds to use in the experiment.
        group (str): The group to use in the experiment.

    Attributes:
        name (str): The name of the experiment.
        agent (Agent): The agent to use in the experiment.
        env (Environment): The environment to use in the experiment.
        env_id (str): The ID of the environment.
        seeds (Tuple[int, ...]): The seeds to use in the experiment.
        group (str): The group to use in the experiment.

    """

    def __init__(
        self,
        name: str,
        agent: Agent,
        env: Environment,
        env_id: str = "",
        seeds: Tuple[int, ...] = (0,),
        group: str = "",
    ):
        self.name = name
        self.agent = agent
        self.env = env
        self.env_id = env_id
        self.seeds = seeds
        self.group = group

    def run(self, do_log: bool = True):
        """Default function to run the experiment. This function compiles the training function, trains the agent, and logs the results.

        Args:
            do_log (bool): Whether to log the results to wandb.
        !!! Warning
            Logging to `wandb` is usually much slower than training the agent itself.
            The time is linear in the number of seeds.

        Returns:
            Tuple: A tuple containing the final training state and the logs.
        """
        print("Running experiment with the following configuration:")
        print(vars(self))
        rng = jnp.asarray([jax.random.PRNGKey(seed) for seed in self.seeds])

        print("Compiling training function...")
        start_time = time.time()
        train_fn = jax.jit(jax.vmap(self.agent.train)).lower(rng).compile()
        compilation_time = time.time() - start_time
        print(f"Compilation time cost: {compilation_time}")

        print("Training agent...")
        start_time = time.time()
        train_state, logs = train_fn(rng)
        training_time = time.time() - start_time
        print(f"Training time cost: {training_time}")

        if not self.agent.hparams.debug and do_log:
            print("Logging final results to wandb...")
            start_time = time.time()
            for seed in self.seeds:
                config = {**vars(self), **asdict(self.agent.hparams)}
                config.update(seed=seed)
                wandb.init(project=self.name, config=config, group=self.group)
                print("Logging results for seed:", seed)
                log = jax.tree.map(lambda x: x[seed], logs)
                self.agent.log_on_train_end(log)
                wandb.finish()
            logging_time = time.time() - start_time
            print(f"Logging time cost: {logging_time}")

        print("Training complete")
        total_time = 0
        print(f"Compilation time cost: {compilation_time}")
        total_time += compilation_time
        print(f"Training time cost: {training_time}")
        total_time += training_time
        if not self.agent.hparams.debug and do_log:
            print(f"Logging time cost: {logging_time}")
            total_time += logging_time
        print(f"Total time cost: {total_time}")
        return train_state, logs

    def run_hparam_search(
        self, hparams_distr: Dict[str, distrax.Distribution], pop_size: int
    ):
        """Function to run a hyperparameter search for the experiment. This function \
        samples hyperparameters from the given distributions, trains the agent, and \
        logs the results.
        
        Args:
            hparams_distr (Dict[str, distrax.Distribution]): A dictionary of \
            hyperparameter distributions. The keys are the hyperparameter names, which \
            must exist in `self.agent.hparams`, and the values are the corresponding \
            distributions.
            pop_size (int): The number of hyperparameter sets to sample.

        Returns:
            Tuple: A tuple containing the final training states and the logs, batched \
            over the hyperparameter sets.
        """
        hparams_fields = fields(self.agent.hparams)
        for k in hparams_distr:
            member = list(filter(lambda x: x.name == k, hparams_fields))
            if (
                len(member) > 0
                and "pytree_node" in member[0].metadata
                and member[0].metadata["pytree_node"] == False
            ):
                raise ValueError(
                    f"Hyperparameter {k} is not a traceable pytree node. "
                    + f"Set pytree_node=True for {k} to include it into the hparam search."
                )

        search_set = []
        for seed in range(pop_size):
            hparams = self.agent.hparams
            key = jax.random.PRNGKey(seed)
            for k, distr in hparams_distr.items():
                hparams = replace(hparams, **{k: distr.sample(seed=key)})
            print("Hparams:", hparams)
            search_set.append(hparams)
        # transpose search set
        len_search_set = len(search_set)
        search_set = jax.tree.map(lambda *x: jnp.stack(x), *search_set)

        rngs = jnp.asarray([jax.random.PRNGKey(seed) for seed in self.seeds])

        def search(hparam_set_sample):
            agent = self.agent.replace(hparams=hparam_set_sample)
            return jax.vmap(agent.train)(rngs)

        print("Running hyperparameter search with the following configuration:")
        print(search_set)

        print("Compiling search function...")
        start_time = time.time()
        search_fn = jax.jit(jax.vmap(search)).lower(search_set).compile()
        compilation_time = time.time() - start_time
        print(f"Compilation time cost: {compilation_time}")

        print("Searching for optimal hyperparameters...")
        start_time = time.time()
        train_states, logs = search_fn(search_set)
        search_time = time.time() - start_time
        print(f"Search time cost: {search_time}")

        print("Logging final results to wandb...")
        start_time = time.time()
        # average over seeds
        for i in range(len_search_set):
            print("Logging results for hparam set:", search_set)
            hparams = jax.tree_map(lambda x: x[i], search_set)
            config = {**vars(self), **asdict(hparams)}
            wandb.init(project=self.name, config=config, group=self.group)
            log = jax.tree_map(lambda x: jnp.mean(x[i], axis=0), logs)
            self.agent.log_on_train_end(log)
            wandb.finish()
        logging_time = time.time() - start_time

        print("Hyperparameter search complete")
        total_time = 0
        print(f"Compilation time cost: {compilation_time}")
        total_time += compilation_time
        print(f"Search time cost: {search_time}")
        total_time += search_time
        print(f"Logging time cost: {logging_time}")
        total_time += logging_time
        print(f"Total time cost: {total_time}")
        return train_states, logs



from __future__ import annotations
from typing import Tuple, TypeVar

import jax
from jax import Array
import jax.numpy as jnp
from flax import struct


from .components import (
    Positionable,
    Directional,
    HasColour,
    HasTag,
    Stochastic,
    Openable,
    Pickable,
    Holder,
    HasSprite,
)
from .rendering.registry import SPRITES_REGISTRY

T = TypeVar("T", bound="Entity")


class Entities(struct.PyTreeNode):
    """Entities enum class to store the names of the entities in the game."""

    WALL: str = struct.field(pytree_node=False, default="wall")
    FLOOR: str = struct.field(pytree_node=False, default="floor")
    PLAYER: str = struct.field(pytree_node=False, default="player")
    GOAL: str = struct.field(pytree_node=False, default="goal")
    KEY: str = struct.field(pytree_node=False, default="key")
    DOOR: str = struct.field(pytree_node=False, default="door")
    LAVA: str = struct.field(pytree_node=False, default="lava")
    BALL: str = struct.field(pytree_node=False, default="ball")
    BOX: str = struct.field(pytree_node=False, default="box")


class EntityIds:
    """EntityIds enum class to store the ids of the entities in the game."""

    UNKNOWN: Array = jnp.asarray(0, dtype=jnp.uint8)
    FLOOR: Array = jnp.asarray(1, dtype=jnp.uint8)
    WALL: Array = jnp.asarray(2, dtype=jnp.uint8)
    DOOR: Array = jnp.asarray(4, dtype=jnp.uint8)
    KEY: Array = jnp.asarray(5, dtype=jnp.uint8)
    BALL: Array = jnp.asarray(6, dtype=jnp.uint8)
    BOX: Array = jnp.asarray(7, dtype=jnp.uint8)
    GOAL: Array = jnp.asarray(8, dtype=jnp.uint8)
    LAVA: Array = jnp.asarray(9, dtype=jnp.uint8)
    PLAYER: Array = jnp.asarray(10, dtype=jnp.uint8)


class Directions:
    """Directions enum class to store the directions in the game."""

    EAST = jnp.asarray(0)
    SOUTH = jnp.asarray(1)
    WEST = jnp.asarray(2)
    NORTH = jnp.asarray(3)


class Entity(Positionable, HasTag, HasSprite):
    """Entities are components that can be placed in the environment, and have a position and a tag.
    To create an entity, use the `create` method."""

    def __getitem__(self: T, idx) -> T:
        return jax.tree_util.tree_map(lambda x: x[idx], self)

    @property
    def name(self) -> str:
        """The name of the entity

        Returns:
            str: the name of the entity"""
        return self.__class__.__name__

    @property
    def shape(self) -> Tuple[int, ...]:
        """The batch shape of the entity. The batch shape is the shape of the entity excluding the dimensions of the component.
        For example, if the entity has a position of shape (batch_size, 2), the shape of the entity is (batch_size,).
        """
        return self.position.shape[:-1]

    @property
    def ndim(self) -> int:
        """The number of dimensions of the entity. The number of dimensions is the number of dimensions of the position minus 1."""
        return self.position.ndim - 1

    @property
    def walkable(self) -> Array:
        """The walkable attribute of the entity. The walkable attribute is a boolean array that indicates if the entity can be walked on."""
        raise NotImplementedError()

    @property
    def transparent(self) -> Array:
        """The transparent attribute of the entity. The transparent attribute is a boolean array that indicates if the entity is transparent to rendering."""
        raise NotImplementedError()


class Wall(Entity, HasColour):
    """Walls are entities that cannot be walked through"""

    @classmethod
    def create(
        cls,
        position: Array,
    ) -> Wall:
        shape = position.shape[:-1]
        grey = jnp.ones(shape, dtype=jnp.uint8) * 5
        return cls(position=position, colour=grey)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.WALL]
        return jnp.broadcast_to(sprite[None], (*self.shape, *sprite.shape))

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(EntityIds.WALL, self.shape)


class Player(Entity, Directional, Holder):
    """Players are entities that can act around the environment"""

    @classmethod
    def create(
        cls,
        position: Array,
        direction: Array,
        pocket: Array,
    ) -> Player:
        return cls(position=position, direction=direction, pocket=pocket)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.PLAYER][self.direction]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # broadcast to batch_size
        return jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(EntityIds.PLAYER, self.shape)


class Goal(Entity, HasColour, Stochastic):
    """Goals are entities that can be reached by the player"""

    @classmethod
    def create(
        cls,
        position: Array,
        probability: Array,
    ) -> Goal:
        shape = position.shape[:-1]
        green = jnp.ones(shape, dtype=jnp.uint8)
        return cls(position=position, probability=probability, colour=green)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.GOAL]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(EntityIds.GOAL, self.shape)


class Key(Entity, Pickable, HasColour):
    """Pickable items are world objects that can be picked up by the player.
    Examples of pickable items are keys, coins, etc."""

    @classmethod
    def create(
        cls,
        position: Array,
        colour: Array,
        id: Array,
    ) -> Key:
        colour = jnp.asarray(colour, dtype=jnp.uint8)
        return cls(position=position, id=id, colour=colour)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.KEY][self.colour]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(EntityIds.KEY, self.shape)


class Door(Entity, Openable, HasColour):
    """Consumable items are world objects that can be consumed by the player.
    Consuming an item requires a tool (e.g. a key to open a door).
    A tool is an id (int) of another item, specified in the `requires` field (-1 if no tool is required).
    After an item is consumed, it is both removed from the `state.entities` collection, and replaced in the grid
    by the item specified in the `replacement` field (0 = floor by default).
    Examples of consumables are doors (to open) food (to eat) and water (to drink), etc.
    """

    @classmethod
    def create(
        cls,
        position: Array,
        requires: Array,
        colour: Array,
        open: Array,
    ) -> Door:
        colour = jnp.asarray(colour, dtype=jnp.uint8)
        return cls(
            position=position,
            requires=requires,
            open=open,
            colour=colour,
        )

    @property
    def walkable(self) -> Array:
        return self.open

    @property
    def transparent(self) -> Array:
        return self.open

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.DOOR][
            self.colour, jnp.asarray(self.open + 2 * self.locked, dtype=jnp.int32)
        ]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(EntityIds.DOOR, self.shape)

    @property
    def locked(self) -> Array:
        return self.requires != jnp.asarray(-1)


class Lava(Entity):
    """Goals are entities that can be reached by the player"""

    @classmethod
    def create(
        cls,
        position: Array,
    ) -> Lava:
        return cls(position=position)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.LAVA]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(EntityIds.LAVA, self.shape)


class Ball(Entity, HasColour, Stochastic):
    """Goals are entities that can be reached by the player"""

    @classmethod
    def create(
        cls,
        position: Array,
        colour: Array,
        probability: Array,
    ) -> Ball:
        return cls(position=position, colour=colour, probability=probability)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.BALL][self.colour]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(EntityIds.BALL, self.shape)


class Box(Entity, HasColour, Holder):
    """Goals are entities that can be reached by the player"""

    @classmethod
    def create(
        cls,
        position: Array,
        colour: Array,
        pocket: Array,
    ) -> Box:
        return cls(position=position, colour=colour, pocket=pocket)

    @property
    def walkable(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(False), self.shape)

    @property
    def transparent(self) -> Array:
        return jnp.broadcast_to(jnp.asarray(True), self.shape)

    @property
    def sprite(self) -> Array:
        sprite = SPRITES_REGISTRY[Entities.BOX][self.colour]
        if sprite.ndim == 3:
            # batch it
            sprite = sprite[None]
        # ensure same batch size
        if sprite.shape[0] != self.position.shape[0]:
            sprite = jnp.broadcast_to(sprite, (*self.shape, *sprite.shape[1:]))
        return sprite

    @property
    def tag(self) -> Array:
        return jnp.broadcast_to(EntityIds.BOX, self.shape)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Union
import jax
import jax.numpy as jnp
from jax import Array
from flax import struct

from .. import rewards, observations, terminations
from ..components import EMPTY_POCKET_ID
from ..rendering.cache import RenderingCache
from ..rendering.registry import PALETTE
from . import Environment
from ..entities import Player, Key, Door, Goal, Wall
from ..states import State
from . import Timestep
from ..grid import mask_by_coordinates, room, random_positions, random_directions
from .registry import register_env


class DoorKey(Environment):
    random_start: bool = struct.field(pytree_node=False, default=False)

    def _reset(self, key: Array, cache: Union[RenderingCache, None] = None) -> Timestep:
        # check minimum height and width
        assert (
            self.height > 3
        ), f"Room height must be greater than 3, got {self.height} instead"
        assert (
            self.width > 4
        ), f"Room width must be greater than 5, got {self.width} instead"

        key, k1, k2, k3, k4 = jax.random.split(key, 5)

        grid = room(height=self.height, width=self.width)

        # door positions
        # col can be between 1 and height - 2
        door_col = jax.random.randint(k4, (), 2, self.width - 2)  # col
        # row can be between 1 and height - 2
        door_row = jax.random.randint(k3, (), 1, self.height - 1)  # row
        door_pos = jnp.asarray((door_row, door_col))
        doors = Door.create(
            position=door_pos,
            requires=jnp.asarray(3),
            open=jnp.asarray(False),
            colour=PALETTE.YELLOW,
        )

        # wall positions
        wall_rows = jnp.arange(1, self.height - 1)
        wall_cols = jnp.asarray([door_col] * (self.height - 2))
        wall_pos = jnp.stack((wall_rows, wall_cols), axis=1)
        # remove wall where the door is
        wall_pos = jnp.delete(
            wall_pos, door_row - 1, axis=0, assume_unique_indices=True
        )
        walls = Wall.create(position=wall_pos)

        # get rooms
        first_room_mask = mask_by_coordinates(
            grid, (jnp.asarray(self.height), door_col), jnp.less
        )
        first_room = jnp.where(first_room_mask, grid, -1)  # put walls where not mask
        second_room_mask = mask_by_coordinates(
            grid, (jnp.asarray(0), door_col), jnp.greater
        )
        second_room = jnp.where(second_room_mask, grid, -1)  # put walls where not mask

        # set player and goal pos
        if self.random_start:
            player_pos = random_positions(k1, first_room)
            player_dir = random_directions(k2)
            goal_pos = random_positions(k2, second_room)
        else:
            player_pos = jnp.asarray([1, 1])
            player_dir = jnp.asarray(0)
            goal_pos = jnp.asarray([self.height - 2, self.width - 2])

        # spawn goal and player
        player = Player.create(
            position=player_pos, direction=player_dir, pocket=EMPTY_POCKET_ID
        )
        goals = Goal.create(position=goal_pos, probability=jnp.asarray(1.0))

        # spawn key
        key_pos = random_positions(k2, first_room, exclude=player_pos)
        keys = Key.create(position=key_pos, id=jnp.asarray(3), colour=PALETTE.YELLOW)

        # remove the wall beneath the door
        grid = grid.at[tuple(door_pos)].set(0)

        entities = {
            "player": player[None],
            "key": keys[None],
            "door": doors[None],
            "goal": goals[None],
            "wall": walls,
        }

        state = State(
            key=key,
            grid=grid,
            cache=cache or RenderingCache.init(grid),
            entities=entities,
        )
        return Timestep(
            t=jnp.asarray(0, dtype=jnp.int32),
            observation=self.observation_fn(state),
            action=jnp.asarray(-1, dtype=jnp.int32),
            reward=jnp.asarray(0.0, dtype=jnp.float32),
            step_type=jnp.asarray(0, dtype=jnp.int32),
            state=state,
        )


register_env(
    "Navix-DoorKey-5x5-v0",
    lambda *args, **kwargs: DoorKey.create(
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        height=5,
        width=5,
        random_start=False,
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-DoorKey-6x6-v0",
    lambda *args, **kwargs: DoorKey.create(
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        height=6,
        width=6,
        random_start=False,
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-DoorKey-8x8-v0",
    lambda *args, **kwargs: DoorKey.create(
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        height=8,
        width=8,
        random_start=False,
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-DoorKey-16x16-v0",
    lambda *args, **kwargs: DoorKey.create(
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        height=16,
        width=16,
        random_start=False,
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-DoorKey-Random-5x5-v0",
    lambda *args, **kwargs: DoorKey.create(
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        height=5,
        width=5,
        random_start=True,
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-DoorKey-Random-6x6-v0",
    lambda *args, **kwargs: DoorKey.create(
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        height=6,
        width=6,
        random_start=True,
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-DoorKey-Random-8x8-v0",
    lambda *args, **kwargs: DoorKey.create(
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        height=8,
        width=8,
        random_start=True,
        *args,
        **kwargs,
    ),
)
register_env(
    "Navix-DoorKey-Random-16x16-v0",
    lambda *args, **kwargs: DoorKey.create(
        observation_fn=kwargs.pop("observation_fn", observations.symbolic),
        reward_fn=kwargs.pop("reward_fn", rewards.on_goal_reached),
        termination_fn=kwargs.pop("termination_fn", terminations.on_goal_reached),
        height=16,
        width=16,
        random_start=True,
        *args,
        **kwargs,
    ),
)



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations
from typing import Tuple


from jax import Array
from flax import struct
import jax.numpy as jnp
import dataclasses


DISCARD_PILE_COORDS = jnp.asarray((0, -1), dtype=jnp.int32)
DISCARD_PILE_IDX = jnp.asarray(-1, dtype=jnp.int32)
EMPTY_POCKET_ID = jnp.asarray(-1, dtype=jnp.int32)
UNSET_DIRECTION = jnp.asarray(-1, dtype=jnp.int32)
UNSET_CONSUMED = jnp.asarray(-1, dtype=jnp.int32)


def field(shape: Tuple[int, ...], **kwargs):
    return dataclasses.field(metadata={"shape": shape}, **kwargs)


class Component(struct.PyTreeNode):
    """Base class for all components in the game.
    Components are used to store the data of the entities in the game."""

    def check_ndim(self, batched: bool = False) -> None:
        return


class Positionable(Component):
    """Flags an entity as positionable in the grid, and provides the `position` attribute"""

    position: Array = field(shape=(2,))
    """The (row, column) position of the entity in the grid as a JAX array, defaults to the discard pile (-1, -1)"""


class Directional(Component):
    """Flags an entity as directional, and provides the `direction` attribute"""

    direction: Array = field(shape=())
    """The direction the entity: 0 = east, 1 = south, 2 = west, 3 = north"""


class HasColour(Component):
    """Flags an entity as having a colour, and provides the `colour` attribute"""

    colour: Array = field(shape=())
    """The colour of the object for rendering. """


class Stochastic(Component):
    """Flags an entity as stochastic, and provides the `probability` attribute

    TODO:
        * consider replace probability (Array) with a distrax.Distribution

    """

    probability: Array = field(shape=())
    """The probability of receiving the reward, if reached."""


class Openable(Component):
    """Flags an entity as openable, and provides the `requires` and `open` attributes"""

    requires: Array = field(shape=())
    """The id of the item required to consume this item. If set, it must be > 0.
    If -1, the door is unlocked and does not require any key to open."""
    open: Array = field(shape=())
    """Open is jnp.asarray(0) if the entity is closed and 1 if open."""


class Pickable(Component):
    """Flags an entity as pickable, and provides the `id` attribute, which is used to identify the item in the inventory"""

    id: Array = field(shape=())
    """The id of the item. If set, it must be >= 1."""


class Holder(Component):
    """Flags an entity as a holder, and provides the `pocket` attribute. The pocket is used to store the id of the item in the pocket."""

    pocket: Array = field(shape=())
    """The id of the item in the pocket (0 if empty)"""


class HasTag(Component):
    """Flags an entity as having a tag, and provides the `tag` attribute. The tag is used to identify the type of the entity in the observations."""

    @property
    def tag(self) -> Array:
        """The tag of the component, used to identify the type of the component in `observations.categorical`"""
        raise NotImplementedError()


class HasSprite(Component):
    """Flags an entity as having a sprite, and provides the `sprite` attribute. The sprite is used to render the entity in the game."""

    @property
    def sprite(self) -> Array:
        raise NotImplementedError()



# Copyright [2023] The Helx Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations
from typing import Tuple

import jax
import jax.numpy as jnp
from jax import Array
from flax import struct

Shape = Tuple[int, ...]


class Space(struct.PyTreeNode):
    """Base class for all spaces in the game. Spaces define the shape and type of the \
        observations, actions and rewards in the game.
        The `sample` method is used to generate random samples from the space.

    !!! note
        To initialize a space, use the `create` method of the specific space class.
        
    TODO: 
        * maximum and minimum should be static objects, not arrays.
    But how do we handle the case when they are not scalars? Maybe numpy arrays?"""

    shape: Shape = struct.field(pytree_node=False)
    dtype: jnp.dtype = struct.field(pytree_node=False)
    minimum: Array
    maximum: Array

    def sample(self, key: Array) -> Array:
        """Generate a random sample from the space.

        Args:
            key (Array): A random key to generate the sample.

        Returns:
            Array: A random sample from the space."""
        raise NotImplementedError()


class Discrete(Space):
    @classmethod
    def create(
        cls, n_elements: int | jax.Array, shape: Shape = (), dtype=jnp.int32
    ) -> Discrete:
        """Create a discrete space with a given number of elements.

        Args:
            n_elements (int | jax.Array): The number of elements in the space.
            shape (Shape): The shape of the space.
            dtype (jnp.dtype): The data type of the space.

        Returns:
            Discrete: A discrete space with the given number of elements."""
        return Discrete(
            shape=shape,
            dtype=dtype,
            minimum=jnp.asarray(0),
            maximum=jnp.asarray(n_elements) - 1,
        )

    def sample(self, key: Array) -> Array:
        """Generate a random sample from the space.

        Args:
            key (Array): A random key to generate the sample.

        Returns:
            Array: A random sample from the space."""
        item = jax.random.randint(key, self.shape, self.minimum, self.maximum)
        # randint cannot draw jnp.uint, so we cast it later
        return jnp.asarray(item, dtype=self.dtype)

    @property
    def n(self) -> Array:
        """The number of elements in the space.

        Returns:
            Array: The number of elements in the space."""
        return self.maximum + 1


class Continuous(Space):
    @classmethod
    def create(
        cls, shape: Shape, minimum: Array, maximum: Array, dtype=jnp.float32
    ) -> Continuous:
        """Create a continuous space with a given shape, minimum and maximum values.

        Args:
            shape (Shape): The shape of the space.
            minimum (Array): The minimum value of the space.
            maximum (Array): The maximum value of the space.
            dtype (jnp.dtype): The data type of the space.

        Returns:
            Continuous: A continuous space with the given shape, minimum and maximum values.
        """
        return Continuous(shape=shape, dtype=dtype, minimum=minimum, maximum=maximum)

    def sample(self, key: Array) -> Array:
        """Generate a random sample from the space.

        Args:
            key (Array): A random key to generate the sample.

        Returns:
            Array: A random sample from the space."""
        assert jnp.issubdtype(self.dtype, jnp.floating)
        # see: https://github.com/google/jax/issues/14003
        lower = jnp.nan_to_num(self.minimum)
        upper = jnp.nan_to_num(self.maximum)
        return jax.random.uniform(
            key, self.shape, minval=lower, maxval=upper, dtype=self.dtype
        )



import time
from timeit import repeat

import jax
import jax.numpy as jnp
import navix as nx


N_TIMEIT_LOOPS = 10
N_REPEAT = 30
N_TIMESTEPS = 10
N_SEEDS = 10000


def test_observation(observation_fn):
    def test(seed):
        env = nx.environments.Room(
            height=5, width=10, max_steps=100, gamma=1.0, observation_fn=observation_fn
        )
        key = jax.random.PRNGKey(seed)
        timestep = env._reset(key)

        # option 1
        # actions = jax.random.randint(key, (100,), 0, 6)
        # timestep = jax.lax.scan(lambda c, x: (env.step(c, x), ()), timestep, actions)[0]

        # option 2
        # for i in range(N_TIMESTEPS):
        #     action = jax.random.randint(key, (), 0, 6)
        #     timestep = env.step(timestep, jnp.asarray(action))

        # option 3
        actions = jax.random.randint(key, (100,), 0, 6)
        jax.lax.while_loop(
            lambda x: x[1] < N_TIMESTEPS,
            lambda x: (env.step(x[0], actions[x[1]]), x[1] + 1),
            (timestep, jnp.asarray(0)),
        )

        return timestep

    # profile navix scanned
    print(
        "Profiling observation {}, N_SEEDS = {}, N_TIMESTEPS = {}".format(
            observation_fn, N_SEEDS, N_TIMESTEPS
        )
    )

    seeds = jnp.arange(N_SEEDS)

    print(f"\tCompiling {observation_fn}...")
    start = time.time()
    test_jit = jax.vmap(test)
    test_jit = jax.jit(test_jit)
    test_jit = test_jit.lower(seeds)
    test_jit = test_jit.compile()
    print("\tCompiled in {:.2f}s".format(time.time() - start))

    print(f"\tRunning {observation_fn}...")
    res = repeat(
        lambda: test_jit(seeds).observation.block_until_ready(),
        number=N_TIMEIT_LOOPS,
        repeat=N_REPEAT,
    )
    res = jnp.asarray(res)
    print(f"\t {jnp.mean(res)} ± {jnp.std(res)}")


if __name__ == "__main__":
    test_observation(nx.observations.none)
    test_observation(nx.observations.categorical)
    test_observation(nx.observations.categorical_first_person)
    test_observation(nx.observations.rgb)
    test_observation(nx.observations.rgb_first_person)



# This implementation of PPO is broadly inspired by:
# https://github.com/luchris429/purejaxrl/blob/main/purejaxrl/ppo.py
# which is in turn inspired by:
# https://github.com/vwxyzjn/cleanrl/blob/master/cleanrl/ppo.py
from functools import partial
import time
from typing import Callable, Dict, Tuple

import distrax
import jax
import jax.numpy as jnp
from jax import Array
import optax
from flax.training.train_state import TrainState
from flax import struct
from flax.linen import FrozenDict as Params
import rlax

from navix.observations import rgb
from navix.agents.agent import Agent, HParams
from navix.environments import Environment
from navix.environments.environment import Timestep
from navix.states import State

from .models import ActorCritic


class PPOHparams(HParams):
    budget: int = struct.field(pytree_node=False, default=1_000_000)
    """Number of environment frames to train for."""
    num_envs: int = struct.field(pytree_node=False, default=16)
    """Number of parallel environments to run."""
    num_steps: int = struct.field(pytree_node=False, default=128)
    """Number of steps to run in each environment per update."""
    num_minibatches: int = struct.field(pytree_node=False, default=8)
    """Number of minibatches to split the data into for training."""
    num_epochs: int = struct.field(pytree_node=False, default=1)
    """Number of epochs to train for."""
    gae_lambda: float = 0.95
    """Lambda parameter of the TD(lambda) return."""
    clip_eps: float = 0.2
    """PPO clip parameter."""
    ent_coef: float = 0.01
    """Entropy coefficient in the total loss."""
    vf_coef: float = 0.5
    """Value function coefficient in the total loss."""
    max_grad_norm: float = 0.5
    """Maximum gradient norm for clipping."""
    lr: float = 2.5e-4
    """Starting learning rate."""
    anneal_lr: bool = struct.field(pytree_node=False, default=True)
    """Whether to anneal the learning rate linearly to 0 at the end of training."""
    normalise_advantage: bool = struct.field(pytree_node=False, default=True)
    """Whether to normalise the advantages in the PPO loss."""
    clip_value_loss: bool = struct.field(pytree_node=False, default=True)
    """Whether to clip the value loss in the PPO loss."""


class Buffer(struct.PyTreeNode):
    done: jax.Array
    action: jax.Array
    reward: jax.Array
    log_prob: jax.Array
    obs: jax.Array
    info: Dict[str, jax.Array]
    t: jax.Array
    state: State


class TrainingState(TrainState):
    env_state: Timestep
    rng: jax.Array
    frames: jax.Array
    epoch: jax.Array
    policy: Callable[[Params, Array], distrax.Distribution] = struct.field(
        pytree_node=False
    )
    value_fn: Callable[[Params, Array], Array] = struct.field(pytree_node=False)


class PPO(Agent):
    hparams: PPOHparams
    network: ActorCritic = struct.field(pytree_node=False)
    env: Environment

    def collect_experience(
        self, train_state: TrainingState
    ) -> Tuple[TrainingState, Buffer]:
        def _env_step(
            collection_state: Tuple[Timestep, jax.Array], _
        ) -> Tuple[Tuple[Timestep, jax.Array], Buffer]:
            env_state, rng = collection_state
            # SELECT ACTION
            rng, _rng = jax.random.split(rng)
            pi = train_state.policy(train_state.params, env_state.observation)
            action = jnp.asarray(pi.sample(seed=_rng))
            log_prob = jnp.asarray(pi.log_prob(action))

            # STEP ENV
            new_env_state = jax.vmap(self.env.step, in_axes=(0, 0))(env_state, action)
            transition = Buffer(
                done=new_env_state.is_done(),  # done(o_{t+1})
                action=action,  # a_t
                reward=new_env_state.reward,  # R(o_t, a_t)
                log_prob=log_prob,  # log π(a_t|o_t)
                obs=env_state.observation,  # o_t
                info=new_env_state.info,  # info(o_{t+1})
                t=env_state.t,  # t
                state=env_state.state,  # s_t
            )
            return (new_env_state, rng), transition

        # collect experience and update env_state
        (env_state, rng), experience = jax.lax.scan(
            _env_step,
            (train_state.env_state, train_state.rng),
            None,
            self.hparams.num_steps,
        )
        train_state = train_state.replace(
            env_state=env_state,
            rng=rng,
            frames=train_state.frames + self.hparams.num_steps * self.hparams.num_envs,
        )
        return train_state, experience

    def evaluate_experience(
        self, train_state: TrainingState, experience: Buffer, last_val: jax.Array
    ) -> Tuple[jax.Array, jax.Array, jax.Array]:
        values = jnp.asarray(
            jax.vmap(train_state.value_fn, in_axes=(None, 0))(
                train_state.params, experience.obs
            )
        )  # (1:T, N)
        adv = jax.vmap(
            rlax.truncated_generalized_advantage_estimation,
            in_axes=(1, 1, None, 1, None),
            out_axes=1,
        )(
            experience.reward,  # (1:T, N)
            (1 - experience.done) * self.env.gamma**experience.t,  # (1:T, N)
            self.hparams.gae_lambda,  # ()
            jnp.concatenate([values, last_val[None]], axis=0),  # (0:T, N)
            True,
        )
        adv = jnp.asarray(adv)  # (0:T, N)
        targets = adv + values
        return values, adv, targets

    def ppo_loss(
        self,
        params: Params,
        transition_batch: Buffer,
        gae: Array,
        targets: Array,
        values_old: Array,
    ):
        # this is already vmapped over the minibatches
        pi, value = jax.vmap(self.network.apply, in_axes=(None, 0))(
            params, transition_batch.obs
        )
        log_prob = pi.log_prob(transition_batch.action)

        # CALCULATE VALUE LOSS
        if self.hparams.clip_value_loss:
            value_loss = jnp.square(value - targets)
            value_clipped = values_old + jnp.clip(
                value - values_old,
                -self.hparams.clip_eps,
                self.hparams.clip_eps,
            )
            value_loss_clipped = 0.5 * jnp.square(value_clipped - targets)
            value_loss = 0.5 * jnp.maximum(value_loss, value_loss_clipped).mean()
        else:
            value_loss = 0.5 * jnp.square(value - targets).mean()

        # CALCULATE ACTOR LOSS
        ratio = jnp.exp(log_prob - transition_batch.log_prob)
        if self.hparams.normalise_advantage:
            gae = (gae - gae.mean()) / (gae.std() + 1e-8)
        loss_actor1 = ratio * gae
        loss_actor2 = (
            jnp.clip(
                ratio,
                1.0 - self.hparams.clip_eps,
                1.0 + self.hparams.clip_eps,
            )
            * gae
        )
        loss_actor = -jnp.minimum(loss_actor1, loss_actor2)
        loss_actor = loss_actor.mean()
        entropy = pi.entropy().mean()

        total_loss = (
            loss_actor
            + self.hparams.vf_coef * value_loss
            - self.hparams.ent_coef * entropy
        )

        # log
        logratio = log_prob - transition_batch.log_prob
        approx_kl = ((ratio - 1) - logratio).mean()
        clipfrac = jnp.mean(jnp.abs(ratio - 1.0) > self.hparams.clip_eps)
        logs = {
            "loss/total_loss": total_loss,
            "loss/value_loss": value_loss,
            "loss/actor_loss": loss_actor,
            "loss/entropy": entropy,
            "loss/approx_kl": approx_kl,
            "loss/clipfrac": clipfrac,
        }
        return total_loss, logs

    def sgd_step(
        self,
        train_state: TrainingState,
        minibatch: Tuple[Buffer, jax.Array, jax.Array, jax.Array],
    ) -> Tuple[TrainingState, Dict]:
        traj_batch, advantages, targets, values_old = minibatch
        grad_fn = jax.value_and_grad(self.ppo_loss, has_aux=True)
        (_, logs), grads = grad_fn(
            train_state.params, traj_batch, advantages, targets, values_old
        )
        train_state = train_state.apply_gradients(grads=grads)
        return train_state, logs

    def update(self, train_state: TrainingState, _) -> Tuple[TrainingState, Dict]:
        # unpack state
        minibatch_size = (
            self.hparams.num_envs
            * self.hparams.num_steps
            // self.hparams.num_minibatches
        )
        # collect experience
        train_state, experience = self.collect_experience(train_state)

        for _ in range(self.hparams.num_epochs):
            # re-evaluate experience at every epoch as per https://arxiv.org/abs/2006.05990
            last_val = train_state.value_fn(
                train_state.params, train_state.env_state.observation
            )  # boostrap
            values, advantages, targets = self.evaluate_experience(
                train_state, experience, last_val
            )

            # Batching and Shuffling
            rng, rng_1 = jax.random.split(train_state.rng)
            train_state = train_state.replace(rng=rng)
            n_samples = minibatch_size * self.hparams.num_minibatches
            assert (
                n_samples == self.hparams.num_steps * self.hparams.num_envs
            ), "batch size must be equal to number of steps * number of envs"
            permutation = jax.random.permutation(rng_1, n_samples)
            samples = (experience, advantages, targets, values)  # (T, N, ...)
            samples = jax.tree_util.tree_map(
                lambda x: x.reshape((n_samples,) + x.shape[2:]), samples
            )  # (T * N, ...)
            shuffled_batch = jax.tree_util.tree_map(
                lambda x: jnp.take(x, permutation, axis=0), samples
            )  # (T * N, ...)

            # One epoch update over all mini-batches
            minibatches = jax.tree_util.tree_map(
                lambda x: jnp.reshape(
                    x, (self.hparams.num_minibatches, -1) + tuple(x.shape[1:])
                ),
                shuffled_batch,
            )
            train_state, logs = jax.lax.scan(self.sgd_step, train_state, minibatches)

        train_state = train_state.replace(
            rng=rng,
            epoch=train_state.epoch + self.hparams.num_epochs,
        )
        logs = jax.tree.map(lambda x: jnp.mean(x), logs)

        learning_rate = train_state.opt_state[1].hyperparams["learning_rate"]  # type: ignore

        # update logs with returns
        logs["done_mask"] = experience.done
        logs["returns"] = experience.info["return"]
        logs["lengths"] = experience.t

        logs["iter/frames"] = train_state.frames
        logs["iter/epochs"] = train_state.epoch
        logs["iter/updates"] = train_state.step
        logs["iter/learning_rate"] = learning_rate

        if self.hparams.log_render:
            b = jax.random.randint(rng, (), 0, self.hparams.num_envs)
            logs["render/human"] = jax.vmap(rgb)(
                jax.tree.map(lambda x: x[:, b], experience.state)
            ).transpose(
                (0, 3, 1, 2)
            )  # (T, 3, H, W)

        # Debugging mode
        if self.hparams.debug:
            jax.debug.callback(self.log, logs, experience)

        return train_state, logs

    def train(self, rng: jax.Array) -> Tuple[TrainingState, Dict]:
        # INIT NETWORK
        rng, _rng = jax.random.split(rng)
        init_x = self.env.observation_space.sample(_rng)
        params = self.network.init(_rng, init_x)

        def linear_schedule(count):
            frac = (
                1.0
                - (count // (self.hparams.num_minibatches * self.hparams.num_epochs))
                / num_updates
            )
            return self.hparams.lr * frac

        lr = linear_schedule if self.hparams.anneal_lr else self.hparams.lr
        tx = optax.chain(
            optax.clip_by_global_norm(self.hparams.max_grad_norm),
            optax.inject_hyperparams(optax.adam)(learning_rate=lr, eps=1e-5),
        )

        # INIT ENV
        rng, _rng = jax.random.split(rng)
        _, reset_rng = jax.lax.scan(
            lambda x, _: (jax.random.split(x)[1], jax.random.split(x)[1]),
            _rng,
            None,
            self.hparams.num_envs,
        )
        reset_rng = jax.random.split(_rng, self.hparams.num_envs)
        env_state = jax.vmap(self.env.reset)(reset_rng)

        # TRAIN LOOP
        num_updates = self.hparams.budget // (
            self.hparams.num_steps * self.hparams.num_envs
        )
        train_state = TrainingState.create(
            apply_fn=jax.vmap(self.network.apply, in_axes=(None, 0)),
            params=params,
            tx=tx,
            env_state=env_state,
            rng=rng,
            frames=jnp.asarray(0, dtype=jnp.int32),
            epoch=jnp.asarray(0, dtype=jnp.int32),
            policy=jax.vmap(
                partial(self.network.apply, method="policy"), in_axes=(None, 0)
            ),
            value_fn=jax.vmap(
                partial(self.network.apply, method="value"), in_axes=(None, 0)
            ),
        )
        start_time = time.time()
        train_state, logs = jax.lax.scan(self.update, train_state, length=num_updates)
        elapsed = time.time() - start_time
        logs["iter/fps"] = jnp.asarray([train_state.frames / elapsed] * num_updates)

        return train_state, logs



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Callable
import difflib


_ENVS_REGISTRY = {}


def registry():
    return _ENVS_REGISTRY


def register_env(name: str, ctor: Callable):
    _ENVS_REGISTRY[name] = ctor


def make(name: str, max_steps: int = 100, **kwargs):
    if name not in registry():
        closest = difflib.get_close_matches(name, registry().keys())
        msg = f"Environment {name} not yet implemented."
        if closest:
            msg += (
                f"Did you mean one of these? {closest}\n"
                + "If not, please open a feature request!"
                + "\nhttps://github.com/epignatelli/navix/issues/new?labels=enhancement"
            )
        raise NotImplementedError(msg)
    ctor = _ENVS_REGISTRY[name]
    return ctor(max_steps=max_steps, **kwargs)


NotImplementedEnvs = [
    "MiniGrid-BlockedUnlockPickup-v0",
    "MiniGrid-LavaCrossingS9N1-v0",
    "MiniGrid-LavaCrossingS9N2-v0",
    "MiniGrid-LavaCrossingS9N3-v0",
    "MiniGrid-LavaCrossingS11N5-v0",
    "MiniGrid-Fetch-5x5-N2-v0",
    "MiniGrid-Fetch-6x6-N2-v0",
    "MiniGrid-Fetch-8x8-N3-v0",
    "MiniGrid-GoToObject-6x6-N2-v0",
    "MiniGrid-GoToObject-8x8-N2-v0",
    "MiniGrid-LockedRoom-v0",
    "MiniGrid-MemoryS17Random-v0",
    "MiniGrid-MemoryS13Random-v0",
    "MiniGrid-MemoryS13-v0",
    "MiniGrid-MemoryS11-v0",
    "MiniGrid-MemoryS9-v0",
    "MiniGrid-MemoryS7-v0",
    "MiniGrid-MultiRoom-N2-S4-v0",
    "MiniGrid-MultiRoom-N4-S5-v0",
    "MiniGrid-MultiRoom-N6-v0",
    "MiniGrid-ObstructedMaze-1Dl-v0",
    "MiniGrid-ObstructedMaze-1Dlh-v0",
    "MiniGrid-ObstructedMaze-1Dlhb-v0",
    "MiniGrid-ObstructedMaze-2Dl-v0",
    "MiniGrid-ObstructedMaze-2Dlh-v0",
    "MiniGrid-ObstructedMaze-2Dlhb-v0",
    "MiniGrid-ObstructedMaze-1Q-v0",
    "MiniGrid-ObstructedMaze-2Q-v0",
    "MiniGrid-ObstructedMaze-Full-v0",
    "MiniGrid-ObstructedMaze-2Dlhb-v1",
    "MiniGrid-ObstructedMaze-1Q-v1",
    "MiniGrid-ObstructedMaze-2Q-v1",
    "MiniGrid-ObstructedMaze-Full-v1",
    "MiniGrid-Playground-v0",
    "MiniGrid-PutNear-6x6-N2-v0",
    "MiniGrid-PutNear-8x8-N3-v0",
    "MiniGrid-RedBlueDoors-6x6-v0",
    "MiniGrid-RedBlueDoors-8x8-v0",
    "MiniGrid-Unlock-v0",
    "MiniGrid-UnlockPickup-v0",
]



import time
from timeit import repeat

import jax
import jax.numpy as jnp
import navix as nx


N_TIMEIT_LOOPS = 10
N_REPEAT = 100
N_TIMESTEPS = 1000
N_SEEDS = 100


def test_observation():
    def test(seed):
        env = nx.environments.Room.create(
            height=10, width=5, max_steps=100, observation_fn=nx.observations.none
        )
        key = jax.random.PRNGKey(seed)
        timestep = env._reset(key)

        actions = jax.random.randint(key, (100,), 0, 6)
        timestep = jax.lax.scan(lambda c, x: (env.step(c, x), ()), timestep, actions)[0]
        return timestep

    # profile navix scanned
    print("Profiling, N_SEEDS = {}, N_TIMESTEPS = {}".format(N_SEEDS, N_TIMESTEPS))

    seeds = jnp.arange(N_SEEDS)

    print(f"\tCompiling {test}...")
    start = time.time()
    test_jit = jax.jit(jax.vmap(test)).lower(seeds).compile()
    print("\tCompiled in {:.2f}s".format(time.time() - start))

    print("\tRunning ...")
    res = repeat(
        lambda: test_jit(seeds).observation.block_until_ready(),
        number=N_TIMEIT_LOOPS,
        repeat=N_REPEAT,
    )
    res = jnp.asarray(res)
    print(f"\t {jnp.mean(res)} ± {jnp.std(res)}")


if __name__ == "__main__":
    test_observation()



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations
from typing import Dict

from jax import Array
import jax.numpy as jnp
from flax import struct


from .components import (
    Positionable,
    HasColour,
)
from .rendering.cache import RenderingCache
from .rendering.registry import PALETTE
from .entities import Entity, Entities, Goal, Wall, Ball, Lava, Key, Door, Box, Player


class EventType:
    """Enumeration of the different types of events that can happen in the environment."""

    NONE: Array = jnp.asarray(-1, dtype=jnp.int32)
    REACH: Array = jnp.asarray(0, dtype=jnp.int32)
    HIT: Array = jnp.asarray(1, dtype=jnp.int32)
    FALL: Array = jnp.asarray(2, dtype=jnp.int32)
    PICKUP: Array = jnp.asarray(3, dtype=jnp.int32)
    OPEN: Array = jnp.asarray(4, dtype=jnp.int32)
    UNLOCK: Array = jnp.asarray(5, dtype=jnp.int32)


class Event(Positionable, HasColour):
    """A struct representing an event that happened in the environment. It contains the
    position of the event, the colour of the entity involved in the event, whether the event
    happened, and the type of event that happened.

    !!! note
        Notice that we need the `happened` property, which flags if an event has
        happened or not, because JAX does not support variable size arrays.
        This means that we cannot add an event to the list in the middle of training.
        Instead, we initialise all events, and mask them out as not happened.


    Attributes:
        position (Array): The (row, column) position of the event in the grid.
        colour (Array): The colour of the entity involved in the event.
        happened (Array): A boolean flag indicating whether the event happened.
        event_type (Array): The type of event that happened."""

    position: Array = jnp.asarray([-1, -1], dtype=jnp.int32)
    colour: Array = PALETTE.UNSET
    happened: Array = jnp.asarray(False, dtype=jnp.bool_)
    event_type: Array = EventType.NONE

    def __eq__(self, other: Event) -> Array:
        return jnp.logical_and(
            jnp.array_equal(self.position, other.position),
            jnp.array_equal(self.colour, other.colour),
        )

    def __ne__(self, other: Event) -> Array:
        return jnp.logical_not(self == other)


class EventsManager(struct.PyTreeNode):
    """A struct that manages the events. It contains the different events that can happen
    in the environment, such as the goal being reached, the player being hit by a ball, etc.

    Attributes:
        goal_reached (Event): An event indicating that the goal has been reached.
        ball_hit (Event): An event indicating that the player has been hit by a ball.
        wall_hit (Event): An event indicating that the player has hit a wall.
        lava_fall (Event): An event indicating that the lava has fallen.
        key_pickup (Event): An event indicating that the player has picked up a key.
        door_opening (Event): An event indicating that the player has opened a door.
        door_unlock (Event): An event indicating that the player has unlocked a door.
        ball_pickup (Event): An event indicating that the player has picked up a ball.
    """

    goal_reached: Event = Event()
    ball_hit: Event = Event()
    wall_hit: Event = Event()
    lava_fall: Event = Event()
    key_pickup: Event = Event()
    door_opening: Event = Event()
    door_unlock: Event = Event()
    ball_pickup: Event = Event()

    def record_walk_into(self, entity: Entity, position: Array) -> EventsManager:
        """Flags an event when the player walks into an entity as happened and returns the
        updated events manager.

        Args:
            entity (Entity): The entity the player walked into.
            position (Array): The position of the entity in the grid.

        Returns:
            EventsManager: The updated events manager."""
        if isinstance(entity, Goal):
            return self.record_goal_reached(entity, position)
        elif isinstance(entity, Wall):
            return self.record_wall_hit(entity, position)
        elif isinstance(entity, Lava):
            return self.record_lava_fall(entity, position)
        return self

    def record_pickup(self, entity: Entity, position: Array) -> EventsManager:
        """Flags an event when the player picks up an entity as happened and returns the
        updated events manager.

        Args:
            entity (Entity): The entity the player picked up.
            position (Array): The position of the entity in the grid.

        Returns:
            EventsManager: The updated events manager."""
        if isinstance(entity, Key):
            return self.record_key_pickup(entity, position)
        elif isinstance(entity, Ball):
            return self.record_ball_pickup(entity, position)
        return self

    def record_goal_reached(self, goal: Goal, position: Array) -> EventsManager:
        """Flags an event when the player reaches the goal as happened and returns the
        updated events manager.

        Args:
            goal (Goal): The goal the player reached.
            position (Array): The position of the goal in the grid.

        Returns:
            EventsManager: The updated events manager."""
        idx = jnp.where(goal.position == position, size=1)[0][0]
        goal = goal[idx]
        return self.replace(
            goal_reached=Event(
                position=position,
                colour=PALETTE.UNSET,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.REACH,
            )
        )

    def record_ball_hit(self, ball: Ball) -> EventsManager:
        """Flags an event when the player is hit by a ball as happened and returns the
        updated events manager.
        
        Args:
            ball (Ball): The ball that hit the player.
        
        Returns:
            EventsManager: The updated events manager."""
        return self.replace(
            ball_hit=Event(
                position=ball.position,
                colour=ball.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.HIT,
            )
        )

    def record_wall_hit(self, wall: Wall, position: Array) -> EventsManager:
        """Flags an event when the player hits a wall as happened and returns the
        updated events manager.
        
        Args:
            wall (Wall): The wall the player hit.
            position (Array): The position of the wall in the grid.
        
        Returns:
            EventsManager: The updated events manager."""
        idx = jnp.where(wall.position == position, size=1)[0][0]
        wall = wall[idx]
        return self.replace(
            wall_hit=Event(
                position=wall.position,
                colour=PALETTE.UNSET,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.HIT,
            )
        )

    def record_grid_hit(self, position: Array) -> EventsManager:
        """Flags an event when the player hits a wall as happened and returns the
        updated events manager.
        
        Args:
            position (Array): The position of the wall in the grid.
        
        Returns:
            EventsManager: The updated events manager."""
        return self.replace(
            wall_hit=Event(
                position=position,
                colour=PALETTE.UNSET,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.HIT,
            )
        )

    def record_lava_fall(self, lava: Lava, position: Array) -> EventsManager:
        """Flags an event when the lava falls as happened and returns the
        updated events manager.
        
        Args:
            lava (Lava): The lava that fell.
            position (Array): The position of the lava in the grid.
        
        Returns:
            EventsManager: The updated events manager."""
        idx = jnp.where(lava.position == position, size=1)[0][0]
        lava = lava[idx]
        return self.replace(
            lava_fall=Event(
                position=lava.position,
                colour=PALETTE.UNSET,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.FALL,
            )
        )

    def record_key_pickup(self, key: Key, position: Array) -> EventsManager:
        """Flags an event when the player picks up a key as happened and returns the
        updated events manager.
        
        Args:
            key (Key): The key the player picked up.
            position (Array): The position of the key in the grid.
        
        Returns:
            EventsManager: The updated events manager."""
        idx = jnp.where(key.position == position, size=1)[0][0]
        key = key[idx]
        return self.replace(
            key_pickup=Event(
                position=key.position,
                colour=key.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.PICKUP,
            )
        )

    def record_door_opening(self, door: Door, position: Array) -> EventsManager:
        """Flags an event when the player opens a door as happened and returns the
        updated events manager.
        
        Args:
            door (Door): The door the player opened.
            position (Array): The position of the door in the grid.
        
        Returns:
            EventsManager: The updated events manager."""
        idx = jnp.where(door.position == position, size=1)[0][0]
        door = door[idx]
        return self.replace(
            door_opening=Event(
                position=door.position,
                colour=door.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.OPEN,
            )
        )

    def record_door_unlock(self, door: Door, position: Array) -> EventsManager:
        """Flags an event when the player unlocks a door as happened and returns the
        updated events manager.

        Args:
            door (Door): The door the player unlocked.
            position (Array): The position of the door in the grid.
        
        Returns:
            EventsManager: The updated events manager."""
        idx = jnp.where(door.position == position, size=1)[0][0]
        door = door[idx]
        return self.replace(
            door_opening=Event(
                position=door.position,
                colour=door.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.UNLOCK,
            )
        )

    def record_ball_pickup(self, ball: Ball, position: Array) -> EventsManager:
        """Flags an event when the player picks up a ball as happened and returns the
        updated events manager.
        
        Args:
            ball (Ball): The ball the player picked up.
            position (Array): The position of the ball in the grid.
        
        Returns:
            EventsManager: The updated events manager."""
        idx = jnp.where(ball.position == position, size=1)[0][0]
        ball = ball[idx]
        return self.replace(
            ball_pickup=Event(
                position=ball.position,
                colour=ball.colour,
                happened=jnp.asarray(True, dtype=jnp.bool_),
                event_type=EventType.PICKUP,
            )
        )


class State(struct.PyTreeNode):
    """The Markovian state of the environment"""

    key: Array
    """The random number generator state"""
    grid: Array
    """The base map of the environment that remains constant throughout the training"""
    cache: RenderingCache
    """The rendering cache to speed up rendering"""
    entities: Dict[str, Entity] = struct.field(default_factory=dict)
    """The entities in the environment, indexed via entity type string representation.
    Batched over the number of entities for each type"""
    events: EventsManager = EventsManager()
    """A struct indicating which events happened this timestep. For example, the
    goal is reached, or the player is hit by a ball."""
    mission: Event | None = None

    def get_entity(self, entity_enum: str) -> Entity:
        """Get an entity from the state by its enum.
        
        Args:
            entity_enum (str): The enum of the entity to get.
        
        Returns:
            Entity: The entity from the state."""
        return self.entities[entity_enum]

    def set_entity(self, entity_enum: str, entity: Entity) -> State:
        """Set an entity in the state by its enum.
        
        Args:
            entity_enum (str): The enum of the entity to set.
            entity (Entity): The entity to set.
        
        Returns:
            State: The updated state."""
        self.entities[entity_enum] = entity
        return self

    def get_walls(self) -> Wall:
        """Gets all the `WALL` entities from the state."""
        return self.entities.get(Entities.WALL, Wall())  # type: ignore

    def set_walls(self, walls: Wall) -> State:
        """Sets the `WALL` entities in the state."""
        self.entities[Entities.WALL] = walls
        return self

    def get_player(self, idx: int = 0) -> Player:
        """Gets the player entity from the state."""
        return self.entities[Entities.PLAYER][idx]  # type: ignore

    def set_player(self, player: Player, idx: int = 0) -> State:
        """Sets the player entity in the state. Notice that we only support one player in the
        environment for now, but this can easily be extended to multiple players."""
        # TODO(epignatelli): this is a hack and won't work in multi-agent settings
        self.entities[Entities.PLAYER] = player[None]
        return self

    def get_goals(self) -> Goal:
        """Gets the goal entity from the state."""
        return self.entities[Entities.GOAL]  # type: ignore

    def set_goals(self, goals: Goal) -> State:
        """Sets the goal entity in the state."""
        self.entities[Entities.GOAL] = goals
        return self

    def get_keys(self) -> Key:
        """Gets the key entity from the state."""
        return self.entities[Entities.KEY]  # type: ignore

    def set_keys(self, keys: Key) -> State:
        """Sets the key entity in the state."""
        self.entities[Entities.KEY] = keys
        return self

    def get_doors(self) -> Door:
        """Gets the door entity from the state."""
        return self.entities[Entities.DOOR]  # type: ignore

    def set_doors(self, doors: Door) -> State:
        """Sets the door entity in the state."""
        self.entities[Entities.DOOR] = doors
        return self

    def get_lavas(self) -> Lava:
        """Gets the lava entity from the state."""
        return self.entities[Entities.LAVA]  # type: ignore

    def get_balls(self) -> Ball:
        """Gets the ball entity from the state."""
        return self.entities[Entities.BALL]  # type: ignore

    def get_boxes(self) -> Ball:
        """Gets the box entity from the state."""
        return self.entities[Entities.BOX]  # type: ignore

    def set_balls(self, balls: Ball) -> State:
        """Sets the ball entity in the state."""
        self.entities[Entities.BALL] = balls
        return self

    def set_boxes(self, boxes: Box) -> State:
        """Sets the box entity in the state."""
        self.entities[Entities.BOX] = boxes
        return self

    def set_events(self, events: EventsManager) -> State:
        """Sets the events in the state."""
        return self.replace(events=events)

    def get_positions(self) -> Array:
        """Get the positions of all the entities in the state."""
        return jnp.concatenate([self.entities[k].position for k in self.entities])

    def get_tags(self) -> Array:
        """Get the tags of all the entities in the state."""
        return jnp.concatenate([self.entities[k].tag for k in self.entities])

    def get_sprites(self) -> Array:
        """Get the sprites of all the entities in the state."""
        return jnp.concatenate([self.entities[k].sprite for k in self.entities])

    def get_transparency(self) -> Array:
        """Get the transparency of all the entities in the state."""
        return jnp.concatenate([self.entities[k].transparent for k in self.entities])



from dataclasses import asdict, dataclass, field
import time
import jax
import tyro
import numpy as np
import jax.numpy as jnp
import wandb
import navix as nx
from navix import observations
from navix.agents import PPO, PPOHparams, ActorCritic
from navix.environments.environment import Environment

# set persistent compilation cache directory
# jax.config.update("jax_compilation_cache_dir", "/tmp/jax-cache/")


@dataclass
class Args:
    project_name = "navix-benchmarks"
    seeds_offset: int = 0
    n_seeds: int = 1
    # env
    env_id: str = "Navix-DoorKey-Random-8x8-v0"
    discount: float = 0.99
    # ppo
    ppo_config: PPOHparams = field(default_factory=PPOHparams)


if __name__ == "__main__":
    args = tyro.cli(Args)

    def FlattenObsWrapper(env: Environment):
        flatten_obs_fn = lambda x: jnp.ravel(env.observation_fn(x))
        flatten_obs_shape = (int(np.prod(env.observation_space.shape)),)
        return env.replace(
            observation_fn=flatten_obs_fn,
            observation_space=env.observation_space.replace(shape=flatten_obs_shape),
        )

    env = nx.make(
        args.env_id,
        observation_fn=observations.symbolic_first_person,
        gamma=args.discount,
    )
    env = FlattenObsWrapper(env)

    ppo_config = args.ppo_config.replace(budget=10_000_000)
    agent = PPO(
        hparams=ppo_config,
        network=ActorCritic(
            action_dim=len(env.action_set),
        ),
        env=env,
    )

    experiment = nx.Experiment(
        name=args.project_name,
        agent=agent,
        env=env,
        env_id=args.env_id,
        seeds=tuple(range(args.seeds_offset, args.seeds_offset + args.n_seeds)),
        group="navix",
    )
    train_state, logs = experiment.run(do_log=False)

    print("Logging final results to wandb...")
    start_time = time.time()
    # average over seeds
    logs_avg = jax.tree.map(lambda x: x.mean(axis=0), logs)
    config = {**vars(experiment), **asdict(agent.hparams)}
    wandb.init(project=experiment.name, config=config, group=experiment.group)
    agent.log_on_train_end(logs_avg)
    wandb.finish()
    logging_time = time.time() - start_time
    print(f"Logging time cost: {logging_time}")



import jax
import jax.numpy as jnp
import navix as nx


def test_grid_from_ascii():
    ascii_map = """########
    #1.....#
    #......#
        #......#
#......#
        #......#
    #......#
    #.....2#
########
########
########
########
    """
    print(ascii_map)

    grid = nx.grid.from_ascii_map(ascii_map)
    print(grid)

    ascii_map = ascii_map.replace("1", "P")
    ascii_map = ascii_map.replace("2", "G")
    grid = nx.grid.from_ascii_map(ascii_map, mapping={"P": 1, "G": 2})
    print(grid)


def test_idx_from_coordinates():
    grid = jnp.zeros((5, 7), dtype=jnp.int32)
    grid = jnp.pad(grid, 1, mode="constant", constant_values=-1)

    positions = jnp.array([[1, 1], [2, 2], [3, 3], [4, 4]])
    indices = nx.grid.idx_from_coordinates(grid, positions)
    positions_after = nx.grid.coordinates_from_idx(grid, indices)
    assert jnp.all(jnp.array_equal(positions, positions_after)), (
        positions,
        positions_after,
    )


def test_random_positions():
    grid = jnp.zeros((5, 7), dtype=jnp.int32)
    grid = jnp.pad(grid, 1, mode="constant", constant_values=-1)

    key = jax.random.PRNGKey(0)
    positions = nx.grid.random_positions(key, grid, n=1)
    assert positions.shape == (2,), positions.shape

    positions = nx.grid.random_positions(key, grid, n=4)
    assert positions.shape == (4, 2), positions.shape

    exclude = jnp.asarray((1, 1))
    positions = nx.grid.random_positions(key, grid, n=50, exclude=exclude)
    for position in positions:
        assert not jnp.array_equal(position, exclude), position
        assert jnp.array_equal(grid[tuple(position)], 0), positions


def test_position_equal():
    # one to one
    a = jnp.array([1, 1])
    b = jnp.array([1, 1])
    assert nx.grid.positions_equal(a, b)
    assert nx.grid.positions_equal(b, a)
    assert not nx.grid.positions_equal(a, b + 1)
    assert not nx.grid.positions_equal(a + 1, b)
    assert not nx.grid.positions_equal(b, a + 1)
    assert not nx.grid.positions_equal(b + 1, a)

    # one to many
    a = jnp.array([1, 1])
    b = jnp.array([[1, 1], [1, 2]])
    assert jnp.array_equal(nx.grid.positions_equal(a, b), jnp.array([True, False]))
    assert jnp.array_equal(nx.grid.positions_equal(b, a), jnp.array([True, False]))
    assert jnp.array_equal(nx.grid.positions_equal(a, b + 1), jnp.array([False, False]))
    assert jnp.array_equal(nx.grid.positions_equal(a + 1, b), jnp.array([False, False]))
    assert jnp.array_equal(nx.grid.positions_equal(b, a + 1), jnp.array([False, False]))
    assert jnp.array_equal(nx.grid.positions_equal(b + 1, a), jnp.array([False, False]))

    # many to many
    a = jnp.array([[1, 1], [1, 2]])
    b = jnp.array([[1, 1], [1, 2]])
    assert jnp.array_equal(nx.grid.positions_equal(a, b), jnp.array([True, True]))
    assert jnp.array_equal(nx.grid.positions_equal(b, a), jnp.array([True, True]))
    assert jnp.array_equal(nx.grid.positions_equal(a, b + 1), jnp.array([False, False]))
    assert jnp.array_equal(nx.grid.positions_equal(a + 1, b), jnp.array([False, False]))
    assert jnp.array_equal(nx.grid.positions_equal(b, a + 1), jnp.array([False, False]))
    assert jnp.array_equal(nx.grid.positions_equal(b + 1, a), jnp.array([False, False]))


if __name__ == "__main__":
    # test_grid_from_ascii()
    # test_idx_from_coordinates()
    # test_random_positions()
    test_position_equal()
    # jax.jit(test_position_equal)()



from typing import Any, Dict, Tuple
import jax
from jax import Array
from flax import struct
from gymnax.environments.environment import (
    Environment as GymnaxEnv,
    EnvParams,
    EnvState,
)
from gymnax.environments.spaces import Discrete as GymnaxDiscrete, Box as GymnaxBox

from .environment import Environment, Timestep


@struct.dataclass
class GymnaxState(EnvState):
    timestep: Timestep
    time: Array


class ToGymnax(GymnaxEnv):
    def __init__(self, env: Environment):
        self.env = env

    @property
    def default_params(self) -> EnvParams:
        return EnvParams(max_steps_in_episode=self.env.max_steps)

    @classmethod
    def wrap(cls, env: Environment) -> Tuple[GymnaxEnv, EnvParams]:
        return cls(env=env), EnvParams(max_steps_in_episode=env.max_steps)

    def action_space(self, params: Any):
        return GymnaxDiscrete(len(self.env.action_set))

    def observation_space(self, params: Any):
        o_space = self.env.observation_space
        return GymnaxBox(
            low=o_space.minimum,
            high=o_space.maximum,
            shape=o_space.shape,
            dtype=o_space.dtype,
        )

    def reset(
        self, key: jax.Array, params: EnvParams | None = None
    ) -> Tuple[Array, EnvState]:
        timestep = self.env.reset(key)
        return (
            timestep.observation,
            GymnaxState(time=timestep.t, timestep=timestep),
        )

    def step(
        self, key: Array, state: GymnaxState, action: jax.Array, params: EnvParams
    ) -> Tuple[Array, EnvState, Array, Array, Dict[str, Any]]:
        timestep = self.env.step(state.timestep, action)
        return (
            timestep.observation,
            GymnaxState(time=timestep.t, timestep=timestep),
            timestep.reward,
            timestep.is_done(),
            timestep.info,
        )



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Callable
from jax import Array
import jax.numpy as jnp

from . import events
from .states import State


def compose(
    *term_functions: Callable[[State, Array, State], Array],
    operator: Callable = jnp.any,
) -> Callable:
    """Compose termination functions into a single termination function.

    Args:
        *term_functions (Callable): List of termination functions.
        operator (Callable): Operator to combine the termination functions.

    Returns:
        Callable: A single termination function."""
    return lambda prev_state, action, state: operator(
        jnp.asarray([term_f(prev_state, action, state) for term_f in term_functions])
    )


def check_truncation(terminated: Array, truncated: Array) -> Array:
    """Check if the episode is truncated or terminated, and returns a value
    that conforms to the `StepType` enum.

    Args:
        terminated (Array): A boolean array indicating whether the episode is terminated.
        truncated (Array): A boolean array indicating whether the episode is truncated.

    Returns:
        Array: An integer array that represents the step type."""
    result = jnp.asarray(truncated + 2 * terminated, dtype=jnp.int32)
    return jnp.clip(result, 0, 2)


def on_goal_reached(prev_state: State, action: Array, state: State) -> Array:
    """Check if the goal has been reached using the `goal_reached` event.

    Args:
        prev_state (State): The previous state of the game.
        action (Array): The action taken by the player.
        state (State): The current state of the game.

    Returns:
        Array: A boolean array indicating whether the goal has been reached."""
    return jnp.asarray(events.on_goal_reached(state), dtype=jnp.bool_)


def on_lava_fall(prev_state: State, action: Array, state: State) -> Array:
    """Check if the lava has fallen using the `lava_fall` event.

    Args:
        prev_state (State): The previous state of the game.
        action (Array): The action taken by the player.
        state (State): The current state of the game.

    Returns:
        Array: A boolean array indicating whether the lava has fallen."""
    return jnp.asarray(events.on_lava_fall(state), dtype=jnp.bool_)


def on_ball_hit(prev_state: State, action: Array, state: State) -> Array:
    """Check if the ball has hit something using the `ball_hit` event.

    Args:
        prev_state (State): The previous state of the game.
        action (Array): The action taken by the player.
        state (State): The current state of the game.

    Returns:
        Array: A boolean array indicating whether the ball has hit something."""
    return jnp.asarray(events.on_ball_hit(state), dtype=jnp.bool_)


def on_door_done(prev_state: State, action: Array, state: State) -> Array:
    """Check if the action `done` has been called in front of a `Door` object with the \
        correct colour.

    Args:
        prev_state (State): The previous state of the game.
        action (Array): The action taken by the player.
        state (State): The current state of the game.

    Returns:
        Array: A boolean array indicating whether the action `done` has been called in \
        front of a `Door` object with the correct colour.
    """
    return jnp.asarray(events.on_door_done(state), dtype=jnp.bool_)


DEFAULT_TERMINATION = compose(on_goal_reached, on_lava_fall, on_ball_hit)



import jax
import jax.numpy as jnp
import navix as nx
import time


N_TIMESTEPS = 10
N_SEEDS = 100


def f(seed):
    key = jax.random.PRNGKey(seed)
    env = nx.environments.Room(16, 16, 8, observation_fn=nx.observations.rgb)
    timestep = env._reset(key)

    for _ in range(N_TIMESTEPS):
        action = jax.random.randint(timestep.state.key, (), 0, 6)
        timestep = env.step(timestep, jnp.asarray(action))
    return timestep


def f_scan(seed):
    key = jax.random.PRNGKey(seed)
    env = nx.environments.Room(16, 16, 8, observation_fn=nx.observations.rgb)
    timestep = env._reset(key)

    def body_fun(carry, x):
        timestep = carry
        action = jax.random.randint(timestep.state.key, (), 0, 6)
        timestep = env.step(timestep, action)
        return timestep, ()

    timestep = jax.lax.scan(
        body_fun,
        timestep,
        [None] * N_TIMESTEPS,
        length=N_TIMESTEPS,
    )[0]
    return timestep


seeds = jnp.arange(N_SEEDS)
function = jax.vmap(f)

# print(f"\tCompiling {function}...")
# start = time.time()
# f_jit = jax.jit(function).lower(seeds).compile()
# print("\tCompiled in {:.2f}s".format(time.time() - start))


with jax.profiler.trace("/tmp/jax-trace", create_perfetto_link=True):
    timestep = function(seeds).observation.block_until_ready()



import jax
import jax.numpy as jnp

import navix as nx
from navix.entities import Directions
from navix.states import State
from navix.components import EMPTY_POCKET_ID


def test_on_navigation_completion():
    grid = jnp.zeros((5, 5), dtype=jnp.int32)

    players = nx.entities.Player(
        position=jnp.asarray((1, 1)), direction=Directions.EAST, pocket=EMPTY_POCKET_ID
    )
    goals = nx.entities.Goal.create(position=jnp.asarray((1, 2)), probability=jnp.asarray(1))
    entities = {
        nx.entities.Entities.PLAYER: players[None],
        nx.entities.Entities.GOAL: goals[None],
    }

    state = State(
        key=jax.random.PRNGKey(0),
        grid=grid,
        cache=nx.rendering.cache.RenderingCache.init(grid),
        entities=entities,
    )
    # should not terminate
    termination = nx.terminations.on_goal_reached(state, jnp.asarray(0), state)
    assert not termination, f"Should not terminate, got {termination} instead"

    # move forward
    new_state = nx.actions.forward(state)
    termination = nx.terminations.on_goal_reached(state, jnp.asarray(0), new_state)
    assert termination, f"Should terminate, got {termination} instead"


def test_check_truncation():
    terminated = jnp.asarray(False)
    truncated = jnp.asarray(False)
    assert nx.terminations.check_truncation(terminated, truncated) == jnp.asarray(
        0, dtype=jnp.int32
    )

    terminated = jnp.asarray(True)
    truncated = jnp.asarray(False)
    assert nx.terminations.check_truncation(terminated, truncated) == jnp.asarray(
        2, dtype=jnp.int32
    )

    terminated = jnp.asarray(False)
    truncated = jnp.asarray(True)
    assert nx.terminations.check_truncation(terminated, truncated) == jnp.asarray(
        1, dtype=jnp.int32
    )

    terminated = jnp.asarray(True)
    truncated = jnp.asarray(True)
    assert nx.terminations.check_truncation(terminated, truncated) == jnp.asarray(
        2, dtype=jnp.int32
    )


if __name__ == "__main__":
    test_on_navigation_completion()
    test_check_truncation()



# Copyright 2023 The Navix Authors.

# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations
from typing import Callable


import jax.numpy as jnp
from jax import Array

from . import events
from .states import State


def compose(
    *reward_functions: Callable[[State, Array, State], Array],
    operator: Callable = jnp.sum,
) -> Callable:
    """Compose multiple reward functions into a single reward function.
    The functions are called in order and the results are reduced using the `operator` \
    function.
    
    Args:
        *reward_functions (Callable[[State, Array, State], Array]): A list of reward functions.
        operator (Callable): The operator to reduce the results of the reward functions.
        It must be a function that takes a list of arrays, or an array and returns an \
        array of size `f32[]`.
    
    Returns:
        Callable: A composed reward function that applies the `operator` to the results of the \
        reward functions."""
    return lambda prev_state, action, state: operator(
        jnp.asarray(
            [f(prev_state, action, state) for f in reward_functions], dtype=jnp.float32
        )
    )


def free(state: State) -> Array:
    """A reward function that always returns 0, to simulate reward-free learning.

    Args:
        state (State): The current state of the game.

    Returns:
        Array: A scalar array `f32[]` with value 0."""
    return jnp.asarray(0.0, dtype=jnp.float32)


def on_goal_reached(prev_state: State, action: Array, state: State) -> Array:
    """A reward function that returns 1 when the goal is reached, and 0 otherwise.

    Args:
        state (State): The current state of the game.

    Returns:
        Array: A scalar array `f32[]` with value 1 if the goal is reached, and 0 otherwise.
    """
    return jnp.asarray(events.on_goal_reached(state), dtype=jnp.float32)


def action_cost(
    prev_state: State, action: Array, new_state: State, cost: float = 0.01
) -> Array:
    """A reward function that returns a negative value when an action is taken. 
    All actions have a cost of `cost`, except for noops.
    
    Args:
        prev_state (State): The previous state of the game.
        action (Array): The action taken.
        new_state (State): The new state of the game.
        cost (float): The cost of taking an action.
    
    Returns:
        Array: A scalar array `f32[]` with value -`cost` if the action is not a noop, \
        and 0 otherwise."""
    # noops are free
    return -jnp.asarray(action != 6, dtype=jnp.float32) * cost


def time_cost(
    prev_state: State, action: Array, new_state: State, cost: float = 0.01
) -> Array:
    """A reward function that returns a negative value as time passes, paying a cost \
    of `cost` at each time step.

    Args:
        prev_state (State): The previous state of the game.
        action (Array): The action taken.
        new_state (State): The new state of the game.
        cost (float): The cost of time passing.
    
    Returns:
        Array: A scalar array `f32[]` with value -`cost`.
    """
    # time always has a cost
    return -jnp.asarray(cost, dtype=jnp.float32)


def wall_hit_cost(
    prev_state: State, action: Array, state: State, cost: float = 0.01
) -> Array:
    """A reward function that returns a negative value when the agent hits a wall, \
    paying a cost of `cost` for each wall hit.
    
    Args:
        state (State): The current state of the game.
        cost (float): The cost of hitting a wall.
    
    Returns:
        Array: A scalar array `f32[]` with value -`cost` if the agent hits a wall, \
        and 0 otherwise."""
    return jnp.asarray(events.on_wall_hit(state), dtype=jnp.float32) * cost


def on_door_done(prev_state: State, action: Array, state: State) -> Array:
    """A reward function that returns a positive value when the agent uses the action \
    `done` in front of a door.
    
    Args:
        state (State): The current state of the game.
        
    Returns:
        Array: A scalar array `f32[]` with value 1 if the agent uses the action `done` in \
        front of a door, and 0 otherwise."""

    return jnp.asarray(events.on_door_done(state), dtype=jnp.float32)


DEFAULT_TASK = compose(on_goal_reached, action_cost)
"""The default task for the game, composed of the `on_goal_reached` and `action_cost` reward functions."""
